# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import pathlib
import time
from contextlib import nullcontext
from typing import TYPE_CHECKING, Any, Dict, Optional

import torch
import torch.nn as nn
import wandb
from huggingface_hub import constants as hf_constants
from megatron_fsdp import MegatronFSDP
from megatron_fsdp.fully_shard import fully_shard_optimizer
from torch.utils.data import DataLoader
from torchao.float8 import precompute_float8_dynamic_scale_for_fsdp
from transformers import AutoProcessor
from transformers.processing_utils import ProcessorMixin
from wandb import Settings

from nemo_automodel._transformers import NeMoAutoModelForImageTextToText, NeMoAutoModelForMultimodalLM
from nemo_automodel._transformers.utils import apply_cache_compatibility_patches
from nemo_automodel.components.checkpoint.checkpointing import Checkpointer, CheckpointingConfig
from nemo_automodel.components.config._arg_parser import parse_args_and_load_config
from nemo_automodel.components.datasets.vlm.collate_fns import COLLATE_FNS
from nemo_automodel.components.distributed.config import MegatronFSDPConfig
from nemo_automodel.components.distributed.cp_utils import make_cp_batch_and_ctx
from nemo_automodel.components.distributed.init_utils import initialize_distributed
from nemo_automodel.components.distributed.pipelining import AutoPipeline
from nemo_automodel.components.distributed.utils import FirstRankPerNode, get_sync_ctx
from nemo_automodel.components.loggers.log_utils import setup_logging
from nemo_automodel.components.loggers.metric_logger import MetricsSample, build_metric_logger
from nemo_automodel.components.loggers.wandb_utils import suppress_wandb_log_messages
from nemo_automodel.components.loss.linear_ce import FusedLinearCrossEntropy
from nemo_automodel.components.loss.masked_ce import MaskedCrossEntropy
from nemo_automodel.components.optim.scheduler import OptimizerParamScheduler
from nemo_automodel.components.quantization.fp8 import build_fp8_config
from nemo_automodel.components.training.rng import ScopedRNG, StatefulRNG
from nemo_automodel.components.training.step_scheduler import StepScheduler
from nemo_automodel.components.training.utils import (
    count_tail_padding,
    prepare_for_final_backward,
    prepare_for_grad_accumulation,
    scale_grads_and_clip_grad_norm,
)
from nemo_automodel.components.utils.compile_utils import build_compile_config
from nemo_automodel.components.utils.model_utils import _supports_logits_to_keep
from nemo_automodel.recipes._dist_setup import setup_distributed
from nemo_automodel.recipes.base_recipe import BaseRecipe

if TYPE_CHECKING:
    from torch.optim import Optimizer

    from nemo_automodel.components.distributed.init_utils import DistInfo

logger = logging.getLogger(__name__)

# ---------------------------
#  Stateless helper functions
# ---------------------------


def _get_model_name(cfg_model):
    if cfg_model.get("pretrained_model_name_or_path", None) is not None:
        return cfg_model.pretrained_model_name_or_path
    elif cfg_model.get("config", None) is not None:
        return cfg_model.config.get("pretrained_model_name_or_path", None)
    else:
        return None


def build_model(
    cfg_model,
    cfg_freeze,
    cfg_peft,
    seed,
    cfg_fp8=None,
    cfg_compile=None,
    device_mesh=None,
    moe_mesh=None,
    distributed_config=None,
    pipeline_config=None,
    cfg_moe=None,
    activation_checkpointing=False,
) -> tuple[nn.Module | AutoPipeline, list["Optimizer"]]:  # noqa: F821
    """Build and initialize a model for VLM.

    Returns:
        The instantiated model and optimizer.
    """
    with ScopedRNG(seed=seed, ranked=True):
        # Build infrastructure kwargs
        kwargs = {
            "peft_config": cfg_peft,
            "device_mesh": device_mesh,
            "moe_mesh": moe_mesh,
            "distributed_config": distributed_config,
            "pipeline_config": pipeline_config,
            "freeze_config": cfg_freeze.to_dict() if cfg_freeze is not None else None,
        }

        if cfg_moe is not None:
            from nemo_automodel.components.moe.config import MoEParallelizerConfig

            if isinstance(cfg_moe, MoEParallelizerConfig):
                kwargs["moe_config"] = cfg_moe
            else:
                moe_dict = cfg_moe.to_dict() if hasattr(cfg_moe, "to_dict") else dict(cfg_moe)
                # activation_checkpointing is handled separately; strip config keys
                moe_dict.pop("activation_checkpointing", None)
                moe_dict.pop("_target_", None)
                kwargs["moe_config"] = MoEParallelizerConfig(**moe_dict)
            kwargs["activation_checkpointing"] = activation_checkpointing

        if cfg_fp8 is not None:
            fp8_config = build_fp8_config(cfg_fp8)
            kwargs["fp8_config"] = fp8_config
        if cfg_compile is not None:
            kwargs["compile_config"] = build_compile_config(cfg_compile)

        # Check if using NeMoAutoModel
        is_nemo_auto_model = cfg_model.get("_target_", None) in (
            NeMoAutoModelForImageTextToText.from_config,
            NeMoAutoModelForImageTextToText.from_pretrained,
            NeMoAutoModelForMultimodalLM.from_config,
            NeMoAutoModelForMultimodalLM.from_pretrained,
        )

        if is_nemo_auto_model:
            # NeMoAutoModel handles infrastructure internally
            model = cfg_model.instantiate(**kwargs)
        else:
            raise ValueError(
                f"VLM finetuning requires NeMoAutoModelForImageTextToText. "
                f"Got model target: {cfg_model.get('_target_', None)}"
            )
    return model


def build_optimizer(model, cfg_opt, distributed_config, device_mesh):
    """Build an optimizer for the model.

    Args:
        model: The model to build an optimizer for.
        cfg_opt: The configuration for the optimizer.
        distributed_config: The distributed configuration.
        device_mesh: The device mesh.
    """
    if device_mesh is not None and "tp" in device_mesh.mesh_dim_names and device_mesh["tp"].size() > 1:
        # TP does not support foreach
        cfg_opt.foreach = False

    optimizer = []
    for part in getattr(model, "parts", [model]):
        trainable_params = list(filter(lambda x: x.requires_grad, part.parameters()))
        assert len(trainable_params) > 0, "trainable_params cannot be empty"
        tmp_optimizer = cfg_opt.instantiate(params=trainable_params)
        if isinstance(distributed_config, MegatronFSDPConfig) and torch.distributed.get_world_size() > 1:
            # Only call fully_shard_optimizer when the model was actually wrapped
            # with MegatronFSDP. When dp_mesh.size()==1 the parallelizer skips
            # MegatronFSDP wrapping and the parameters won't carry the required
            # _megatron_fsdp_model attribute.
            if isinstance(part, MegatronFSDP):
                fully_shard_optimizer(tmp_optimizer)
        optimizer.append(tmp_optimizer)

    return optimizer


def build_checkpoint_config(cfg_ckpt, cache_dir, model_repo_id, is_peft) -> CheckpointingConfig:
    """Build a checkpoint configuration.

    Args:
        cfg_ckpt: Configuration for checkpointing.
        cache_dir: Cache directory for the model.
        model_repo_id: Model repository ID.
        is_peft: Whether the model is PEFT.

    Returns:
        The instantiated checkpoint configuration.
    """
    ckpt_kwargs = dict(
        enabled=True,
        checkpoint_dir="checkpoints/",
        model_save_format="safetensors",
        model_repo_id=model_repo_id,
        model_cache_dir=cache_dir if cache_dir is not None else hf_constants.HF_HUB_CACHE,
        save_consolidated=True,
        is_peft=is_peft,
    )
    if cfg_ckpt is not None:
        cfg_ckpt = cfg_ckpt.to_dict()
        cfg_ckpt.pop("restore_from", None)
        ckpt_kwargs |= cfg_ckpt
    if ckpt_kwargs.get("is_peft", False) and ckpt_kwargs.get("model_save_format") == "torch_save":
        raise ValueError(
            "PEFT checkpointing is not supported for torch_save format. Save using `safetensors` format instead."
        )
    checkpoint_config = CheckpointingConfig(**ckpt_kwargs)
    return checkpoint_config


def build_loss_fn(cfg_loss):
    """Build a loss function.

    Args:
        cfg_loss: Loss function configuration.

    Returns:
        The instantiated loss function.
    """
    return cfg_loss.instantiate()


def build_dataloader(
    cfg_ds, cfg_dl, pretrained_model_name_or_path, cfg_processor, device_mesh, seed, local_batch_size
) -> tuple[DataLoader, ProcessorMixin]:
    """Build a DataLoader for the VLM dataset.

    Args:
        cfg_ds: Dataset configuration.
        cfg_dl: DataLoader configuration.
        pretrained_model_name_or_path: Pretrained model name or path for processor loading.
        cfg_processor: Processor configuration or None.
        device_mesh: Device mesh for distributed training.
        seed: Random seed.
        local_batch_size: Local batch size.

    Returns:
        The instantiated DataLoader and processor.
    """
    dist_sampler_kwargs = {
        "shuffle": cfg_dl.get("shuffle", True),
    }
    if device_mesh is not None:
        dist_sampler_kwargs |= {
            "num_replicas": device_mesh["dp"].size(),
            "rank": device_mesh["dp"].get_local_rank(),
        }

    with ScopedRNG(seed=seed, ranked=True):
        processor = None
        processor_kwargs = {}

        with FirstRankPerNode():
            if cfg_processor is not None and hasattr(cfg_processor, "instantiate"):
                processor = cfg_processor.instantiate()
            elif cfg_processor is not None:
                processor_kwargs = cfg_processor.to_dict()

            # If no processor was instantiated, try AutoProcessor
            if processor is None:
                try:
                    processor = AutoProcessor.from_pretrained(pretrained_model_name_or_path, **processor_kwargs)
                except Exception as e:
                    # Some models do not provide an AutoProcessor
                    processor = None
                    logging.warning(f"AutoProcessor not available for {pretrained_model_name_or_path} ({e}). ")

            ds = cfg_ds.instantiate(path_or_dataset=cfg_ds.path_or_dataset)

        sampler = torch.utils.data.distributed.DistributedSampler(
            ds,
            **dist_sampler_kwargs,
        )
        collate_cfg = cfg_dl.get("collate_fn", None)
        if collate_cfg:
            collate_fn = lambda examples: collate_cfg.instantiate(examples=examples, processor=processor)
        else:
            processor_type = type(processor).__name__
            if processor_type not in COLLATE_FNS:
                processor_type = "default"
                logging.warning(f"You are using {processor_type} with default collate function.")
            collate_fn = lambda examples: COLLATE_FNS[processor_type](examples, processor)

        return cfg_dl.instantiate(
            dataset=ds, sampler=sampler, collate_fn=collate_fn, batch_size=local_batch_size
        ), processor


def build_distributed(cfg_dist: Dict[str, Any]) -> "DistInfo":  # noqa: F821
    """Build and initialize distributed training resources.

    Args:
        cfg_dist: Configuration for distributed training.

    Returns:
        Distributed training information from initialize_distributed.
    """
    backend = cfg_dist.get("backend", "nccl")
    timeout = cfg_dist.get("timeout_minutes", 1)
    return initialize_distributed(backend=backend, timeout_minutes=timeout)


def build_step_scheduler(cfg, dataloader, dp_group_size, local_batch_size):
    """Build the step scheduler.

    Args:
        cfg: configuration for the StepScheduler class.
        dataloader: the training dataloader, used for extracting the epoch_len (in batches).
        dp_group_size: the size of the data parallel group.
        micro_batch_size: the size of the micro batch.

    Returns:
        StepScheduler: the configured StepScheduler.
    """
    assert "_target_" not in cfg, "_target_ not permitted in step scheduler"
    default_kwargs = dict(
        num_epochs=10,
        global_batch_size=32,
        local_batch_size=local_batch_size,
        dp_size=dp_group_size,
        ckpt_every_steps=100,
        dataloader=dataloader,
    )
    if cfg is not None:
        default_kwargs |= cfg.to_dict()
    return StepScheduler(**default_kwargs)


def build_lr_scheduler(cfg, optimizer, step_scheduler) -> list[OptimizerParamScheduler] | None:  # noqa: F821
    """Build the learning rate scheduler.

    Args:
        cfg: Configuration for the OptimizerParamScheduler.
        optimizer: The optimizer to be scheduled.
        step_scheduler: The step scheduler to extract training parameters.

    Returns:
        OptimizerParamScheduler: The configured learning rate scheduler, or None if not configured.
    """
    if cfg is None:
        return None

    # Calculate total steps for the training run
    total_epochs = step_scheduler.num_epochs
    epoch_len = len(step_scheduler.dataloader)
    grad_acc_steps = step_scheduler.grad_acc_steps

    # Total optimizer steps (accounting for gradient accumulation)
    total_steps = (total_epochs * epoch_len) // grad_acc_steps
    if step_scheduler.max_steps is not None:
        total_steps = min(total_steps, step_scheduler.max_steps)

    optimizer_param_schedulers = []
    user_kwargs = cfg.to_dict()
    default_kwargs = dict(
        lr_warmup_steps=min(1000, total_steps // 10),  # 10% warmup or max 1000 steps
        lr_decay_steps=total_steps,
        lr_decay_style="cosine",
        wd_incr_steps=total_steps,
        wd_incr_style="constant",
    )

    if not isinstance(optimizer, list):
        optimizer = [optimizer]

    for opt in optimizer:
        base_lr = opt.param_groups[0]["lr"]
        default_kwargs.update(
            dict(
                optimizer=opt,
                init_lr=base_lr * 0.1,  # Start warmup at 10% of base LR
                max_lr=base_lr,
                min_lr=base_lr * 0.01,  # End at 1% of base LR
                start_wd=opt.param_groups[0].get("weight_decay", 0.0),
                end_wd=opt.param_groups[0].get("weight_decay", 0.0),
            )
        )
        default_kwargs.update(user_kwargs)
        optimizer_param_schedulers.append(OptimizerParamScheduler(**default_kwargs))

    logger.info(
        f"Building LR scheduler with total_steps={total_steps}, "
        f"warmup_steps={default_kwargs['lr_warmup_steps']}, "
        f"decay_style={default_kwargs['lr_decay_style']}"
    )

    return optimizer_param_schedulers


def build_wandb(cfg) -> wandb.Run:
    """Instantiates wandb and returns the instance. If no name is given, it will use the model name.

    Args:
        cfg: Configuration for wandb.

    Returns:
        The wandb instance.
    """
    assert cfg.get("wandb", None) is not None
    kwargs = cfg.wandb.to_dict()
    if kwargs.get("name", "") == "":
        kwargs["name"] = "_".join(_get_model_name(cfg.model).split("/")[-2:])
    run = wandb.init(
        **kwargs,
        config=cfg.to_dict(),
        settings=Settings(silent=True),
    )
    return run


def calculate_loss(loss_fn, **kwargs) -> torch.Tensor:
    """Calculate the loss.

    Args:
        loss_fn: Loss function.
        **kwargs: Keyword arguments for the loss function.

    Returns:
        The loss.
    """
    loss_fn_kwargs = {"num_label_tokens": kwargs.pop("num_label_tokens", None)}
    if isinstance(loss_fn, FusedLinearCrossEntropy):
        model = kwargs.pop("model")
        labels = kwargs.pop("labels")

        # find the lm_head in the model
        lm_head = None
        if hasattr(model, "get_output_embeddings"):
            lm_head = model.get_output_embeddings().weight
        else:
            for n, p in model.named_parameters(remove_duplicate=False):
                if "lm_head" in n and n.endswith(".weight"):
                    lm_head = p
                    break
        if lm_head is None:
            raise ValueError("lm_head.weight not found in model")

        # unshard the possibly sharded lm_head
        lm_head = lm_head.full_tensor() if hasattr(lm_head, "full_tensor") else lm_head
        loss_fn_kwargs.update(
            {
                "hidden_states": kwargs.pop("hidden_states"),
                "labels": labels,
                "lm_weight": lm_head,
            }
        )
    else:
        loss_fn_kwargs.update(
            {
                "logits": kwargs.pop("logits"),
                "labels": kwargs.pop("labels"),
            }
        )

    return loss_fn(**loss_fn_kwargs)


# ---------------------------------------------------------------------------
#  Trainer class – orchestration only
# ---------------------------------------------------------------------------


class FinetuneRecipeForVLM(BaseRecipe):
    """Recipe for fine-tuning a VLM model."""

    def __init__(self, cfg):
        """Initialize the recipe with configuration.

        Args:
            cfg: Configuration dictionary/object for training.
        """
        self.cfg = cfg

    # ------------------ build phase ------------------
    def setup(self):
        """Builds all components needed for training/validation/logging/checkpointing/etc.

        This is the last place where self.cfg should be referenced.

        Raises:
            NotImplemented: Raises if it tries to restore a checkpoint; will be removed.
        """
        torch.cuda.reset_peak_memory_stats()
        self.dist_env = build_distributed(self.cfg.get("dist_env", {}))
        setup_logging()

        apply_cache_compatibility_patches()

        # Set up the stateful random number generator
        self.rng = StatefulRNG(seed=self.cfg.get("seed", 42), ranked=True)

        self.dist_setup = setup_distributed(self.cfg, world_size=self.dist_env.world_size)
        self.distributed_config = self.dist_setup.strategy_config
        self.device_mesh = self.dist_setup.device_mesh
        self.moe_mesh = self.dist_setup.moe_mesh
        self.pp_enabled = self.dist_setup.pp_enabled
        self.pipeline_config = self.dist_setup.pipeline_config

        if self.dist_env.is_main and hasattr(self.cfg, "wandb"):
            suppress_wandb_log_messages()
            run = build_wandb(self.cfg)
            logging.info("🚀 View run at {}".format(run.url))

        # Log experiment details on main rank
        self._log_experiment_details()
        self._log_library_versions()

        # Build loss_fn (will be set on pipeline_config if PP enabled)
        self.loss_fn = build_loss_fn(self.cfg.loss_fn)

        # Pipeline runtime fields: override pp_batch_size and pp_microbatch_size
        if self.pp_enabled:
            pp_batch_size = self.cfg.step_scheduler.local_batch_size
            pp_microbatch_size = self.cfg.get("distributed.pipeline.pp_microbatch_size", 1)

            assert pp_batch_size // pp_microbatch_size >= self.dist_setup.pp_size, (
                f"pp_batch_size {pp_batch_size} // pp_microbatch_size {pp_microbatch_size} must be >= pp_size {self.dist_setup.pp_size}"
            )

            assert not isinstance(self.distributed_config, MegatronFSDPConfig), (
                "MegatronFSDPConfig is not supported when pipeline parallelism is enabled"
            )

            # Update pipeline_config runtime fields
            self.pipeline_config.pp_batch_size = pp_batch_size
            self.pipeline_config.pp_microbatch_size = pp_microbatch_size
            self.pipeline_config.patch_stage_backward_maybe_with_nosync = self.cfg.get(
                "model.backend.enable_fsdp_optimizations", False
            )
            self.pipeline_config.loss_fn = self.loss_fn

        # Build components with VLM-specific functions
        self.peft_config = None
        if self.cfg.get("peft", None) is not None:
            self.peft_config = self.cfg.peft.instantiate()

        # Build checkpoint config
        checkpoint_config = build_checkpoint_config(
            self.cfg.get("checkpoint", None),
            self.cfg.get("model.cache_dir", None),
            _get_model_name(self.cfg.model),
            True if self.cfg.get("peft", None) else False,
        )

        if self.cfg.get("clip_grad_norm.max_norm", None) is not None:
            self.max_grad_norm = float(self.cfg.clip_grad_norm.max_norm)
        else:
            logging.info("No clip_grad_norm.max_norm specified in config, using default value of 1.0")
            self.max_grad_norm = 1.0

        # Create Checkpointer instance
        self.checkpointer = Checkpointer(
            config=checkpoint_config,
            dp_rank=self._get_dp_rank(include_cp=True),
            tp_rank=self._get_tp_rank(),
            pp_rank=self._get_pp_rank(),
            moe_mesh=self.moe_mesh,
        )

        model = build_model(
            self.cfg.model,
            self.cfg.get("freeze_config", None),
            self.peft_config,
            seed=self.cfg.get("seed", 42),
            cfg_fp8=self.cfg.get("fp8", None),
            cfg_compile=self.cfg.get("compile", None),
            device_mesh=self.device_mesh,
            moe_mesh=self.moe_mesh,
            distributed_config=self.distributed_config,
            pipeline_config=self.pipeline_config,
            cfg_moe=self.dist_setup.moe_config,
            activation_checkpointing=self.dist_setup.activation_checkpointing,
        )
        self.optimizer = build_optimizer(model, self.cfg.optimizer, self.distributed_config, self.device_mesh)

        if not _supports_logits_to_keep(model) and not isinstance(self.loss_fn, MaskedCrossEntropy):
            logger.warning("logits_to_keep not found in model.forward. Using MaskedCrossEntropy instead.")
            self.loss_fn = MaskedCrossEntropy()

        if isinstance(model, AutoPipeline):
            self.model_parts = model.parts
            self.pp = model
        else:
            self.model_parts = [model]
            self.pp = None

        self.dataloader, self.processor = build_dataloader(
            self.cfg.dataset,
            self.cfg.dataloader,
            _get_model_name(self.cfg.model),
            self.cfg.get("processor", None),
            device_mesh=self.device_mesh,
            seed=self.cfg.get("seed", 42),
            local_batch_size=self.cfg.get("step_scheduler.local_batch_size", 1),
        )

        # Build validation dataloader if the config provides it
        self.val_dataloader = None
        if "validation_dataset" in self.cfg:
            self.val_dataloader, _ = build_dataloader(
                self.cfg.validation_dataset,
                self.cfg.validation_dataloader,
                _get_model_name(self.cfg.model),
                self.cfg.get("processor", None),
                device_mesh=self.device_mesh,
                seed=self.cfg.get("seed", 42),
                local_batch_size=self.cfg.get("step_scheduler.local_batch_size", 1),
            )

        self.best_metric_key = self.cfg.get("checkpoint.best_metric_key", "default")
        # Scheduler
        self.step_scheduler = build_step_scheduler(
            self.cfg.get("step_scheduler", None),
            self.dataloader,
            self._get_dp_group_size(),
            local_batch_size=self.cfg.get("step_scheduler.local_batch_size", 1),
        )

        # Build learning rate scheduler
        self.lr_scheduler = build_lr_scheduler(self.cfg.get("lr_scheduler", None), self.optimizer, self.step_scheduler)

        # Log model, parameter counts, norms, optimizer and scheduler
        self._log_model_and_optimizer_details(self.model_parts, self.optimizer, self.lr_scheduler)

        restore_from = self.cfg.get("checkpoint.restore_from", None)

        # Initialize JSONL loggers
        self.metric_logger_train = build_metric_logger(
            pathlib.Path(self.checkpointer.config.checkpoint_dir) / "training.jsonl"
        )
        self.metric_logger_valid = build_metric_logger(
            pathlib.Path(self.checkpointer.config.checkpoint_dir) / "validation.jsonl"
        )

        # Optionally resume
        self.load_checkpoint(restore_from)

        # Log step scheduler details
        self._log_step_scheduler_details(self.step_scheduler)

    # ------------------ main loop ------------------
    def run_train_validation_loop(self):
        """Run the training loop over all epochs and batches.

        For each batch, perform a forward pass, compute loss, backpropagate,
        and update model parameters when necessary. Also prints loss every gradient step.
        """
        for mp in self.model_parts:
            mp.train()
        self.timestamp = time.perf_counter()

        for epoch in self.step_scheduler.epochs:
            self.step_scheduler.set_epoch(epoch)
            for batch_idx, batches in enumerate(self.step_scheduler):
                log_data = self._run_train_optim_step(batches, self.max_grad_norm)
                # log
                self.log_train_metrics(log_data)

                val_loss = {}
                if self.step_scheduler.is_val_step and self.val_dataloader is not None:
                    if self.pp_enabled:
                        logger.warning("Validation is not supported for pipeline parallelism")
                    else:
                        val_log_data = self._run_validation_epoch(self.val_dataloader)
                        val_loss["val_loss"] = val_log_data.metrics["val_loss"]
                        self.log_val_metrics(val_log_data)
                    for mp in self.model_parts:
                        mp.train()

                if self.step_scheduler.is_ckpt_step:
                    self.save_checkpoint(
                        epoch,
                        self.step_scheduler.step,
                        log_data.metrics["loss"],
                        val_loss,
                        best_metric_key=self.best_metric_key,
                    )

        # Close JSONL loggers after training loop completes
        self.metric_logger_train.close()
        self.metric_logger_valid.close()

        self.checkpointer.close()

    # ------------------ helpers ------------------
    def _forward_backward_step(
        self,
        idx,
        batch,
        *,
        loss_buffer,
        num_label_tokens,
        num_batches,
        is_train: bool = True,
    ):
        batch = {
            k: (
                {dk: dv.to(self.dist_env.device, non_blocking=True) if dv is not None else None for dk, dv in v.items()}
                if isinstance(v, dict)
                else (v.to(self.dist_env.device, non_blocking=True) if isinstance(v, torch.Tensor) else v)
            )
            for k, v in batch.items()
        }

        train_ctx, batch = make_cp_batch_and_ctx(self.device_mesh, batch)
        labels = batch.pop("labels")

        if self.pp_enabled:
            if not is_train:
                logging.info("Skipping forward pass for validation because pipeline parallelism is enabled")
                return

            with train_ctx():
                losses = [] if self.pp.info.has_last_stage else None
                if self.pp.info.has_last_stage:
                    masked_labels = labels.clone()
                    targets = masked_labels
                else:
                    targets = None

                input_ids = batch.pop("input_ids")

                # VLM: Custom chunking for pixel_values and image_grid
                # These tensors have non-standard structure that can't be naively chunked by dim 0
                # TODO: @HuiyingLi properly handle pixel_values split
                pixel_values = batch.pop("pixel_values", None)
                image_grid_hws = batch.pop("image_grid_hws", None)
                image_grid_thw = batch.pop("image_grid_thw", None)

                image_grid = image_grid_hws if image_grid_hws is not None else image_grid_thw

                if self.pp.info.has_first_stage and pixel_values is not None and image_grid is not None:
                    stage0_model = self.model_parts[0]
                    n_microbatches = self.pp._info.schedule._n_microbatches
                    batch_size = input_ids.shape[0]
                    n_images = image_grid.shape[0]

                    patch_counts = image_grid.prod(dim=1)
                    cumsum = torch.cumsum(patch_counts, dim=0)

                    pixel_values_chunks = []
                    image_grid_chunks = []

                    if n_images == batch_size:
                        # 1 image per sample
                        images_per_mb = batch_size // n_microbatches
                        for mb_idx in range(n_microbatches):
                            img_start = mb_idx * images_per_mb
                            img_end = min(img_start + images_per_mb, n_images)

                            image_grid_chunks.append(image_grid[img_start:img_end])

                            patch_start = 0 if img_start == 0 else cumsum[img_start - 1].item()
                            patch_end = cumsum[img_end - 1].item() if img_end > 0 else 0
                            pixel_values_chunks.append(pixel_values[int(patch_start) : int(patch_end)])
                    else:
                        pixel_values_chunks.append(pixel_values)
                        image_grid_chunks.append(image_grid)
                        for _ in range(n_microbatches - 1):
                            pixel_values_chunks.append(pixel_values[:0])
                            image_grid_chunks.append(image_grid[:0])
                        logging.warning(
                            f"VLM chunking: n_images={n_images} != batch_size={batch_size}, giving all images to first microbatch"
                        )

                    # Store pre-chunked tensors on model for forward to use
                    stage0_model._vlm_pixel_values_chunks = pixel_values_chunks
                    stage0_model._vlm_image_grid_hws_chunks = image_grid_chunks
                    stage0_model._vlm_chunk_idx = 0

                if self.pp.info.has_first_stage:
                    self.pp.info.schedule.step(input_ids, target=targets, losses=losses, **batch)
                else:
                    self.pp.info.schedule.step(target=targets, losses=losses, **batch)

                # Clear stored VLM chunks after PP step
                if self.pp.info.has_first_stage and pixel_values is not None and image_grid is not None:
                    stage0_model = self.model_parts[0]
                    stage0_model._vlm_pixel_values_chunks = None
                    stage0_model._vlm_image_grid_hws_chunks = None
                    stage0_model._vlm_chunk_idx = None

            if self.pp.info.has_last_stage:
                local_loss = torch.sum(torch.stack(losses))
            else:
                local_loss = torch.tensor(0.0, device=self.dist_env.device)

            loss_buffer.append(local_loss.clone().detach())
        else:
            model = self.model_parts[0]
            sync_ctx = (
                get_sync_ctx(
                    model,
                    idx == num_batches - 1,
                    defer_fsdp_grad_sync=getattr(self.distributed_config, "defer_fsdp_grad_sync", True),
                )
                if is_train
                else nullcontext()
            )
            with train_ctx(), sync_ctx:
                if isinstance(self.loss_fn, FusedLinearCrossEntropy):
                    # use num_logits_to_keep to avoid full logits matrix in memory
                    out = model(logits_to_keep=1, **batch)
                    if "hidden_states" not in out:
                        raise ValueError(
                            "FusedLinearCrossEntropy requires the model to output hidden states. "
                            "Set `model.text_config.output_hidden_states=True` in the config."
                        )
                else:
                    out = model(**batch)

                local_loss = calculate_loss(
                    self.loss_fn,
                    logits=getattr(out, "logits", out),
                    labels=labels,
                    model=model,
                    hidden_states=out.hidden_states[-1] if getattr(out, "hidden_states", None) is not None else None,
                    num_label_tokens=num_label_tokens,
                )
                loss_buffer.append(local_loss.clone().detach())
                if is_train:
                    (local_loss * self._get_dp_group_size(include_cp=True)).backward()

    def _run_train_optim_step(self, batches, max_grad_norm: Optional[float] = None):
        """Execute a single training step.

        Args:
            batches: List of batches of training data.
            max_grad_norm: Gradient clipping norm. Optional, if None will not clip gradients.
        """
        num_label_tokens = torch.tensor(
            sum((batch["labels"] != -100).sum().item() for batch in batches), dtype=torch.long
        )
        num_label_tokens = self._dp_allreduce(num_label_tokens).item()
        loss_buffer = []

        # number of tokens in the batch, excluding any tail padding.
        num_tokens_in_batch = torch.tensor(
            sum(batch["labels"].numel() - count_tail_padding(batch["labels"]) for batch in batches),
            dtype=torch.long,
        )
        num_tokens_in_batch = self._dp_allreduce(num_tokens_in_batch).item()

        num_batches = len(batches)
        prepare_for_grad_accumulation(self.model_parts, pp_enabled=self.pp_enabled)

        for i, batch in enumerate(batches):
            if i == num_batches - 1:
                prepare_for_final_backward(self.model_parts, pp_enabled=self.pp_enabled)

            self._forward_backward_step(
                i, batch, loss_buffer=loss_buffer, num_label_tokens=num_label_tokens, num_batches=num_batches
            )

        grad_norm = scale_grads_and_clip_grad_norm(
            max_grad_norm=max_grad_norm,
            model_parts=self.model_parts,
            norm_type=2.0,
            pp_enabled=self.pp_enabled,
            device_mesh=self.device_mesh,
            moe_mesh=self.moe_mesh,
            ep_axis_name="ep" if self.moe_mesh is not None and "ep" in self.moe_mesh.mesh_dim_names else None,
            pp_axis_name="pp" if self.pp_enabled else None,
            foreach=True,
            num_label_tokens=num_label_tokens,
            dp_group_size=self._get_dp_group_size(include_cp=True),
        )

        # Note(MegatronFSDP): Need to call these functions for MegatronFSDP if not using latest api
        # self.model.finish_grad_sync()

        self.checkpointer.maybe_wait_for_staging()
        for opt in self.optimizer:
            opt.step()
            opt.zero_grad(set_to_none=True)

        if hasattr(self.model_parts[0], "update_moe_gate_bias"):
            for mp in self.model_parts:
                mp.update_moe_gate_bias()

        if self.lr_scheduler is not None:
            for scheduler in self.lr_scheduler:
                scheduler.step(1)

        # Precompute FP8 scales
        fp8_config = self.cfg.get("fp8", None)
        if (
            fp8_config is not None
            and fp8_config.get("enabled", False)
            and fp8_config.get("precompute_float8_dynamic_scale_for_fsdp", False)
            and self.device_mesh is not None
            and self.device_mesh["dp_shard"].size() > 1
        ):
            precompute_float8_dynamic_scale_for_fsdp(self.model_parts[0])

        # Note(MegatronFSDP): Need to call these functions for MegatronFSDP if not using latest api
        # self.model.install_optimized_model_weights()
        # self.model.zero_grad_buffer()

        t = time.perf_counter()
        time_delta = t - self.timestamp
        self.timestamp = t
        tps = num_tokens_in_batch / time_delta
        reporting_loss = torch.sum(torch.stack(loss_buffer))
        reporting_loss = self._dp_allreduce(reporting_loss, include_cp=True)
        if self.pp_enabled:
            # PP uses sum reduction per microbatch (no internal normalization).
            # Divide by num_label_tokens to get the mean loss, same as non-PP.
            reporting_loss = reporting_loss / num_label_tokens
            reporting_loss = reporting_loss.float().to(self.dist_env.device)
            # Send loss to first rank from the last PP stage of rank0's mesh coords.
            # This avoids picking a global-rank sender from a different EP/PP group.
            if self.device_mesh is not None and "pp" in self.device_mesh.mesh_dim_names:
                dim_names = list(self.device_mesh.mesh_dim_names)
                mesh = self.device_mesh.mesh
                idx = []
                for name in dim_names:
                    if name == "pp":
                        idx.append(-1)
                    else:
                        idx.append(0)
                src_rank = mesh[tuple(idx)].item()
            else:
                src_rank = self.device_mesh.mesh.reshape(-1)[-1].item()
            if self.dist_env.rank == src_rank:
                torch.distributed.send(reporting_loss, dst=0)
            elif self.dist_env.is_main:
                torch.distributed.recv(reporting_loss, src=src_rank)

        reporting_loss = reporting_loss.item()
        # fix reporting_loss, tps across ranks

        return MetricsSample(
            step=self.step_scheduler.step,
            epoch=self.step_scheduler.epoch,
            metrics={
                "loss": reporting_loss,
                "grad_norm": grad_norm,
                "lr": self.optimizer[0].param_groups[0]["lr"],
                "mem": torch.cuda.max_memory_allocated() / 1024**3,
                "tps": tps,
                "tps_per_gpu": tps / self._get_cp_group_size() / max(self._get_dp_group_size(), 1),
                "num_tokens_per_step": num_tokens_in_batch,
                "num_label_tokens": num_label_tokens,
            },
        )

    @torch.no_grad()
    def _run_validation_epoch(self, val_dataloader):
        """Run one pass over `self.val_dataloader`."""
        with ScopedRNG(seed=1, ranked=True):
            for mp in self.model_parts:
                mp.eval()

            total_loss = 0.0
            total_tokens = 0
            total_num_label_tokens = 0
            for batch in val_dataloader:
                batch = {k: v.to(self.dist_env.device, non_blocking=True) for k, v in batch.items()}
                labels = batch.pop("labels")
                num_label_tokens = (labels != -100).sum().item()

                if (
                    self.device_mesh
                    and "position_ids" not in batch
                    and (self.device_mesh["cp"].size() > 1 or self.device_mesh["tp"].size() > 1)
                ):
                    batch["position_ids"] = (
                        torch.arange(0, batch["input_ids"].shape[1]).unsqueeze(0).to(self.model_parts[0].device)
                    )

                train_ctx, batch = make_cp_batch_and_ctx(self.device_mesh, batch, labels)
                with train_ctx():
                    if isinstance(self.loss_fn, FusedLinearCrossEntropy):
                        out = self.model_parts[0](logits_to_keep=1, **batch)
                    else:
                        out = self.model_parts[0](**batch)
                    local_loss = calculate_loss(
                        self.loss_fn,
                        logits=getattr(out, "logits", out),
                        labels=labels,
                        model=self.model_parts[0],
                        hidden_states=out.hidden_states[-1]
                        if getattr(out, "hidden_states", None) is not None
                        else None,
                        num_label_tokens=num_label_tokens,
                    )
                    total_num_label_tokens += num_label_tokens

                total_loss += local_loss.item() * num_label_tokens
                total_tokens += num_label_tokens

        # Aggregate across ranks if distributed is initialized
        total_loss = self._dp_allreduce(torch.FloatTensor([total_loss]), include_cp=True).item()
        total_tokens = self._dp_allreduce(torch.LongTensor([total_tokens]), include_cp=True).item()
        total_num_label_tokens = self._dp_allreduce(torch.LongTensor([total_num_label_tokens])).item()

        val_loss = total_loss / max(total_tokens, 1e-8)

        return MetricsSample(
            step=self.step_scheduler.step,
            epoch=self.step_scheduler.epoch,
            metrics={
                "val_loss": val_loss,
                "lr": self.optimizer[0].param_groups[0]["lr"],
                "num_label_tokens": total_num_label_tokens,
                "mem": torch.cuda.max_memory_allocated() / 1024**3,
            },
        )

    def log_val_metrics(self, log_data):
        """Log metrics to wandb and other loggers
        Args:
            log_data: MetricsSample object, containing:
                step: int, the current step.
                epoch: int, the current epoch.
                metrics: Dict[str, float], containing:
                    "val_loss": Validation loss.
                    "lr": Learning rate.
                    "num_label_tokens": Number of label tokens.
                    "mem": Memory allocated.
        """

        if not self.dist_env.is_main or log_data is None:
            return

        if wandb.run is not None:
            wandb.log(log_data.to_dict(), step=log_data.step)

        # JSONL validation log
        self.metric_logger_valid.log(log_data)

        logging.info(
            "[val] step {} | epoch {} | loss {:.4f} | lr {:.2e} | num_label_tokens {}".format(
                log_data.step,
                log_data.epoch,
                log_data.metrics["val_loss"],
                log_data.metrics["lr"],
                log_data.metrics["num_label_tokens"],
            )
        )

    def log_train_metrics(self, log_data) -> float:
        """Log metrics to wandb.

        Args:
            train_loss: Training loss.
            grad_norm: Grad norm from the training step.
            num_tokens_in_batch: Total number of loss tokens.
            tps: Tokens per second.
        """
        if not self.dist_env.is_main:
            return

        # Log to remote services (WandB) according to step_scheduler frequency
        if self.step_scheduler.is_remote_logging_step:
            if wandb.run is not None:
                wandb.log(log_data.to_dict(), step=self.step_scheduler.step)

        # JSONL training log (always log for detailed local records)
        self.metric_logger_train.log(log_data)
        logging.info(
            "step {} | epoch {} | loss {:.4f} | grad_norm {:.4f} | lr {:.2e} | mem {:.2f} GiB | tps {:.2f}({:.2f}/gpu) | num_label_tokens {}".format(
                log_data.step,
                log_data.epoch,
                log_data.metrics["loss"],
                log_data.metrics["grad_norm"],
                log_data.metrics["lr"],
                log_data.metrics["mem"],
                log_data.metrics["tps"],
                log_data.metrics["tps_per_gpu"],
                log_data.metrics["num_label_tokens"],
            )
        )
        torch.cuda.reset_peak_memory_stats()


# ---------------------------------------------------------------------------
# Entry point
# ---------------------------------------------------------------------------


def main(config_path=None):
    """Main entry point for the fine-tuning recipe.

    Loads the configuration, sets up the trainer, and initiates the training loop.
    """
    if config_path is None:
        config_path = pathlib.Path(__file__).parent.resolve() / "gemma3" / "gemma3_vl_4b_cord_v2.yaml"
    cfg = parse_args_and_load_config(config_path)
    trainer = FinetuneRecipeForVLM(cfg)
    trainer.setup()
    trainer.run_train_validation_loop()


if __name__ == "__main__":
    main()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tensor utilities for device transfers and memory management in distributed settings.

This module provides utilities for handling tensor operations across different devices
and distributed tensor types, with optimizations for performance in distributed training scenarios.
"""

from typing import Iterable, Union

import torch
from torch.distributed.tensor import DTensor


@torch.no_grad()
def get_cpu_state_dict(
    state_generator: Iterable[tuple[str, Union[torch.Tensor, DTensor]]],
    pin_memory: bool = False,
) -> dict[str, torch.Tensor]:
    """Copy the state dict generator to CPU memory.

    Args:
        state_generator (Iterable[tuple[str, Union[torch.Tensor, DTensor]]]):
            An iterable that yields (key, tensor) pairs from a model state.
        pin_memory (bool, optional):
            Whether to allocate the CPU tensors in pinned memory for faster GPU transfer.
            Defaults to False.

    Returns:
        dict[str, torch.Tensor]: A dictionary mapping parameter names to CPU tensors.
    """
    new_state_dict = {}
    for k, v in state_generator:
        val = to_local_if_dtensor(v)

        if len(val.shape) == 0:
            new_state_dict[k] = val.cpu()
        else:
            cpu_tensor = torch.empty(*val.shape, device="cpu", pin_memory=pin_memory, dtype=val.dtype)
            cpu_tensor.copy_(val, non_blocking=True)
            new_state_dict[k] = cpu_tensor

    torch.cuda.synchronize()
    return new_state_dict


def to_cpu(v):
    """
    Move a tensor or distributed tensor to the CPU.

    This function takes an input tensor, which can be either a `DTensor` (distributed tensor)
    or a standard `Tensor`, and ensures that it is moved to the CPU.

    Args:
        v (DTensor | Tensor | any): The input value, which can be a `DTensor`, `Tensor`, or
                                    any other object. If `DTensor`, it checks the device and
                                    moves the tensor accordingly.

    Returns:
        Tensor | any: The corresponding CPU tensor if `v` is a `DTensor` or `Tensor`,
                    otherwise returns `v` unchanged.

    Raises:
        ValueError: If `v` is a `DTensor` but its device is neither 'cuda' nor 'cpu'.

    Example:
        >>> t = torch.tensor([1, 2, 3], device='cuda')
        >>> to_cpu(t)  # Moves tensor to CPU
        tensor([1, 2, 3])

        >>> dt = DTensor(torch.tensor([4, 5, 6], device='cuda'))
        >>> to_cpu(dt)  # Moves DTensor to CPU
        tensor([4, 5, 6])
    """
    if isinstance(v, DTensor):
        if v.device.type == "cuda":
            return v.full_tensor().cpu()
        elif v.device.type == "cpu":
            return v._local_tensor
        else:
            raise ValueError("Unknown device " + str(v.device))
    elif isinstance(v, torch.Tensor):
        return v.cpu()
    else:
        return v


def to_local_if_dtensor(tensor: Union[torch.Tensor, DTensor]) -> torch.Tensor:
    """Returns the local shard of the given tensor if it is a DTensor.

    Taken and modified from: https://github.com/NVIDIA/Megatron-LM/blob/605f618f237cda8fa80132bc2ccff933512d5a0d/megatron/core/utils.py#L746
    """
    with torch.no_grad():
        return tensor.to_local() if isinstance(tensor, DTensor) else tensor



# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for working with model output objects.

HuggingFace `ModelOutput` types typically store `hidden_states` as a tuple of tensors
(`tuple[Tensor, ...]`) when `output_hidden_states=True`.

However, some custom models may store the *final* hidden state tensor directly in the
`hidden_states` field (i.e., a single `[B, T, H]` tensor) to reduce memory/overhead.

Downstream training code should be robust to both representations.
"""

from __future__ import annotations

from typing import Any, Optional

import torch

try:
    # DTensor is not a torch.Tensor subclass; treat it as tensor-like for extraction.
    from torch.distributed.tensor import DTensor  # type: ignore

    _TENSOR_LIKE = (torch.Tensor, DTensor)
except Exception:  # pragma: no cover
    DTensor = None  # type: ignore[assignment]
    _TENSOR_LIKE = (torch.Tensor,)


def get_final_hidden_states(model_output: Any) -> Optional[Any]:
    """Return the final hidden-states tensor from a HF-like model output.

    Supports both common layouts:
    - `hidden_states` is a tuple/list of tensors (HF default) → return last non-None entry
    - `hidden_states` is a single tensor-like object → return it as-is

    Args:
        model_output: A HF `ModelOutput`-like object, or a plain dict with a `hidden_states` key.

    Returns:
        The final hidden states tensor-like object, or None if not present.
    """
    if model_output is None:
        return None

    if isinstance(model_output, dict):
        hidden_states = model_output.get("hidden_states", None)
    else:
        hidden_states = getattr(model_output, "hidden_states", None)

    if hidden_states is None:
        return None

    if isinstance(hidden_states, _TENSOR_LIKE):
        return hidden_states

    if isinstance(hidden_states, (tuple, list)):
        # Some models may include None entries; select the last non-None entry.
        for hs in reversed(hidden_states):
            if hs is None:
                continue
            if not isinstance(hs, _TENSOR_LIKE):
                raise TypeError(f"Expected hidden_states entries to be tensor-like, got {type(hs)}")
            return hs
        return None

    raise TypeError(
        f"Unexpected hidden_states type {type(hidden_states)}; expected a tensor-like object or a tuple/list of them."
    )



# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
import math
import re
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Optional

import torch
from transformers import GptOssConfig

from nemo_automodel.components.checkpoint.state_dict_adapter import StateDictAdapter
from nemo_automodel.components.models.common import BackendConfig
from nemo_automodel.components.moe.config import MoEConfig

if TYPE_CHECKING:
    from torch.distributed.device_mesh import DeviceMesh

FP4_VALUES = [
    +0.0,
    +0.5,
    +1.0,
    +1.5,
    +2.0,
    +3.0,
    +4.0,
    +6.0,
    -0.0,
    -0.5,
    -1.0,
    -1.5,
    -2.0,
    -3.0,
    -4.0,
    -6.0,
]


class GPTOSSStateDictAdapter(StateDictAdapter):
    def __init__(
        self,
        config: GptOssConfig,
        moe_config: MoEConfig,
        backend: BackendConfig,
        dtype: torch.dtype = torch.bfloat16,
    ):
        self.config = config
        self.moe_config = moe_config
        self.backend = backend
        self.dtype = dtype
        self._uses_model_prefix = True

        # Key mapping from HF GPT OSS format to internal format
        self.hf_to_internal_map = {
            # Router mapping
            "mlp.router.weight": "mlp.gate.weight",
            "mlp.router.bias": "mlp.gate.bias",
            "mlp.experts.gate_up_proj": "mlp.experts.gate_and_up_projs",
            "mlp.experts.down_proj": "mlp.experts.down_projs",
        }
        if self.backend.attn == "te":
            self.hf_to_internal_map["self_attn.sinks"] = "self_attn.attn_module.softmax_offset"

        # Reverse mapping for to_hf conversion
        self.internal_to_hf_map = {v: k for k, v in self.hf_to_internal_map.items() if v is not None}

    # replace _apply_key_mapping with leaf-aware replacement
    def _apply_key_mapping(self, state_dict: dict[str, Any], mapping: dict[str, str]) -> dict[str, Any]:
        keys_to_remove = set()
        new_state_dict = {}
        for key, value in state_dict.items():
            new_key = key
            for pattern, replacement in mapping.items():
                if replacement is not None and key.endswith(pattern):
                    new_key = key[: -len(pattern)] + replacement
                    break
            new_state_dict[new_key] = value
            keys_to_remove.add(key)

        for key in keys_to_remove:
            del state_dict[key]
        return new_state_dict

    def _dequantize_block_scale_tensors(self, state_dict: dict[str, Any]) -> dict[str, Any]:
        layer_name_to_quantized_weights = defaultdict(dict)

        # create the mapping from layer name to quantized weights {layer_name: {"blocks"/"scales": value}}
        for key, value in list(state_dict.items()):
            if key.endswith("_blocks") or key.endswith("_scales"):
                layer_name, quantized_name = key.rsplit("_", 1)
                layer_name_to_quantized_weights[layer_name][quantized_name] = value
                del state_dict[key]

        # dequantize the experts
        for layer_name, quantized_dict in layer_name_to_quantized_weights.items():
            dequantized_weights = self._convert_moe_packed_tensors(quantized_dict["blocks"], quantized_dict["scales"])
            state_dict[layer_name] = dequantized_weights

        # clean up the memory
        torch.cuda.empty_cache()
        gc.collect()
        return state_dict

    def _convert_moe_packed_tensors(
        self,
        blocks,
        scales,
        dtype: torch.dtype = torch.bfloat16,
        rows_per_chunk: int = 32768 * 1024,
    ) -> torch.Tensor:
        """
        Convert the mxfp4 weights to bfloat16.

        Source: https://github.com/huggingface/transformers/blob/869735d37d0f929311ac6611728c482a4414ba8c/src/transformers/integrations/mxfp4.py#L77
        """
        # Check if blocks and scales are on CPU, and move to GPU if so
        if not blocks.is_cuda and torch.cuda.is_available() and torch.distributed.get_world_size() > 1:
            blocks = blocks.cuda()
            scales = scales.cuda()

        scales = scales.to(torch.int32) - 127  # that's because 128=2**7

        assert blocks.shape[:-1] == scales.shape, f"{blocks.shape[:-1]=} does not match {scales.shape=}"

        lut = torch.tensor(FP4_VALUES, dtype=dtype, device=blocks.device)

        *prefix_shape, G, B = blocks.shape
        rows_total = math.prod(prefix_shape) * G

        blocks = blocks.reshape(rows_total, B)
        scales = scales.reshape(rows_total, 1)

        if isinstance(blocks, torch.distributed.tensor.DTensor):
            out = torch.distributed.tensor.empty(
                (rows_total, B * 2), placements=blocks.placements, device_mesh=blocks.device_mesh, dtype=dtype
            )
        else:
            out = torch.empty((rows_total, B * 2), dtype=dtype, device=blocks.device)

        for r0 in range(0, rows_total, rows_per_chunk):
            r1 = min(r0 + rows_per_chunk, rows_total)

            blk = blocks[r0:r1]
            exp = scales[r0:r1]
            sub = out[r0:r1]

            # Work on local shards to avoid DTensor advanced indexing
            blk_local = blk.to_local() if hasattr(blk, "to_local") else blk
            sub_local = sub.to_local() if hasattr(sub, "to_local") else sub
            exp_local = exp.to_local() if hasattr(exp, "to_local") else exp

            # Ensure uint8 for nibble extraction
            blk_local = blk_local.to(torch.uint8)

            # nibble indices -> int64 (local)
            idx_lo_local = (blk_local & 0x0F).to(torch.long)
            idx_hi_local = (blk_local >> 4).to(torch.long)

            sub_local[:, 0::2] = lut[idx_lo_local]
            sub_local[:, 1::2] = lut[idx_hi_local]

            torch.ldexp(sub_local, exp_local, out=sub_local)
            del idx_lo_local, idx_hi_local, blk_local, exp_local, sub_local, blk, exp, sub

        out = out.reshape(*prefix_shape, G, B * 2).view(*prefix_shape, G * B * 2)
        del blocks, scales, lut

        # Final logical layout is (n_experts, 2880, hidden_dim) after transpose.
        out = out.transpose(1, 2).contiguous()
        # Restore desired DTensor sharding: shard experts (dim 0) by 'ep' and hidden dim (dim 2) by 'ep_shard'.
        if isinstance(out, torch.distributed.tensor.DTensor):
            placements = []
            mesh_dim_names = out.device_mesh.mesh_dim_names
            for dim_name in mesh_dim_names:
                if dim_name == "ep":
                    placements.append(torch.distributed.tensor.Shard(0))
                elif dim_name == "ep_shard":
                    placements.append(torch.distributed.tensor.Shard(2))
                else:
                    raise ValueError(f"Unexpected dimension name: {dim_name}")
            if placements != out.placements:
                out = out.redistribute(placements=tuple(placements))
        return out

    def to_hf(
        self, state_dict: dict[str, Any], exclude_key_regex: Optional[str] = None, quantization: bool = False, **kwargs
    ) -> dict[str, Any]:
        """Convert from native model state dict to HuggingFace format."""
        hf_state_dict = {}
        for fqn, tensor in state_dict.items():
            converted_tensors = self.convert_single_tensor_to_hf(
                fqn, tensor, exclude_key_regex=exclude_key_regex, quantization=quantization, **kwargs
            )
            for key, value in converted_tensors:
                hf_state_dict[key] = value

        return hf_state_dict

    def from_hf(
        self,
        hf_state_dict: dict[str, Any],
        device_mesh: Optional["DeviceMesh"] = None,
        **kwargs,
    ) -> dict[str, Any]:
        """Convert HF checkpoint to native format.
        - Apply key mappings from HF to internal format
        - Add quantization block and scale tensors
        """
        # Detect model prefix usage
        for key in hf_state_dict.keys():
            if key.startswith("model."):
                self._uses_model_prefix = True
                break

        native_state_dict = dict(hf_state_dict)
        native_state_dict = self._dequantize_block_scale_tensors(native_state_dict)
        native_state_dict = self._apply_key_mapping(native_state_dict, self.hf_to_internal_map)

        return native_state_dict

    def convert_single_tensor_to_hf(self, fqn: str, tensor: Any, **kwargs) -> list[tuple[str, Any]]:
        """Convert a single tensor from native format to HuggingFace format.

        Args:
            fqn: Fully qualified name of the tensor in native format
            tensor: The tensor to convert
            **kwargs: Additional arguments for conversion

        Returns:
            List of (fqn, tensor) tuples in HuggingFace format
        """
        quantization = kwargs.get("quantization", False)
        exclude_key_regex = kwargs.get("exclude_key_regex", None)

        hf_fqn = fqn
        for pattern, replacement in self.internal_to_hf_map.items():
            if fqn.endswith(pattern):
                hf_fqn = fqn[: -len(pattern)] + replacement
                break

        if exclude_key_regex:
            if re.match(exclude_key_regex, hf_fqn):
                return []

        if quantization:
            if hf_fqn.endswith("gate_up_proj") or hf_fqn.endswith("down_proj"):
                layer_name, projection_type = hf_fqn.rsplit(".", 1)
                n_experts, _, dim = tensor.shape

                if isinstance(tensor, torch.distributed.tensor.DTensor):
                    device_mesh = tensor.device_mesh
                    # Ensure quantized tensors shard only along dim 0 for safe flattening in conversion
                    orig_placements = tensor.placements
                    safe_placements = []
                    found_shard_dim0 = False
                    for p in orig_placements:
                        if isinstance(p, torch.distributed.tensor.Shard):
                            if p.dim == 0 and not found_shard_dim0:
                                safe_placements.append(p)
                                found_shard_dim0 = True
                            else:
                                safe_placements.append(torch.distributed.tensor.Replicate())
                        else:
                            safe_placements.append(p)
                    blocks_tensors = torch.distributed.tensor.ones(
                        (n_experts, dim, 90, 16),
                        placements=tuple(safe_placements),
                        device_mesh=device_mesh,
                        dtype=torch.uint8,
                    )
                    scales_tensors = torch.distributed.tensor.ones(
                        (n_experts, dim, 90),
                        placements=tuple(safe_placements),
                        device_mesh=device_mesh,
                        dtype=torch.uint8,
                    )
                else:
                    blocks_tensors = torch.ones((n_experts, dim, 90, 16), dtype=torch.uint8)
                    scales_tensors = torch.ones((n_experts, dim, 90), dtype=torch.uint8)

                return [
                    (f"{layer_name}.{projection_type}_blocks", blocks_tensors),
                    (f"{layer_name}.{projection_type}_scales", scales_tensors),
                ]

        return [(hf_fqn, tensor)]



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributed.tensor import DTensor


class MaskedCrossEntropy(nn.Module):
    def __init__(self, fp32_upcast: bool = True, ignore_index: int = -100, reduction: str = "sum"):
        """
        Masked cross-entropy loss.

        Args:
            fp32_upcast (bool): if True it will cast logits to float32 before computing
                cross entropy. Default: True.
            ignore_index (int): label to ignore in CE calculation. Defaults to -100.
            reduction (str): type of reduction. Defaults to "sum".
        """
        super().__init__()
        self.fp32_upcast = fp32_upcast
        self.ignore_index = ignore_index
        self.reduction = reduction

    def forward(
        self,
        logits: torch.Tensor,
        labels: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
        num_label_tokens: Optional[int] = None,
    ) -> torch.Tensor:
        """
        Compute the masked cross-entropy loss between logits and targets.

        If a mask is provided, the loss is computed per element, multiplied by the mask,
        and then averaged. If no mask is provided, the standard cross-entropy loss is used.

        Args:
            logits (torch.Tensor): The predicted logits with shape [batch_size, seq_len, vocab_size] where C is the number of classes.
            labels (torch.Tensor): The ground truth class indices with shape [batch_size, seq_len].
            mask (torch.Tensor, optional): A tensor that masks the loss computation. Items marked with
                1 will be used to calculate loss, otherwise ignored. Must be broadcastable to the shape
                of the loss. Defaults to None.

        Returns:
            torch.Tensor: The computed loss as a scalar tensor.
        """
        # this may happen with CPUOffloadPolicy
        if labels.device != logits.device:
            labels = labels.to(logits.device)  # pragma: no cover
        # reshape to (N, C) and (N,) respectively
        logits = logits.view(-1, logits.size(-1))
        labels = labels.view(-1)
        if mask is not None:
            with torch.no_grad():
                if mask.device != labels.device:
                    mask = mask.to(labels.device)  # pragma: no cover
                labels.masked_fill_(mask.view(-1) == 0, self.ignore_index)
                del mask
        if self.fp32_upcast:
            logits = logits.float()

        if isinstance(logits, DTensor):
            logits = logits.full_tensor()

        if isinstance(labels, DTensor):
            labels = labels.full_tensor()

        loss = F.cross_entropy(logits, labels, reduction=self.reduction)
        if num_label_tokens is not None:
            assert self.reduction == "sum", "num_label_tokens is only supported when reduction is 'sum'"
            loss = loss / num_label_tokens
        return loss



# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""State dict adapter for Step3p5 model.

Step3p5 uses grouped MoELinear weights with shape [n_exp, out, in], different from
the standard per-expert format. This adapter handles conversion between:

HF Format (Step3p5):
    model.layers.{L}.moe.gate_proj.weight    # [n_exp, inter, dim]
    model.layers.{L}.moe.up_proj.weight      # [n_exp, inter, dim]
    model.layers.{L}.moe.down_proj.weight    # [n_exp, dim, inter]
    model.layers.{L}.moe.gate.weight         # [n_exp, dim] (router)
    model.layers.{L}.moe.router_bias         # [n_exp] (router bias, optional)
    model.layers.{L}.share_expert.*.weight   # Shared expert

Native Format (Automodel):
    model.layers.{L}.moe.experts.gate_and_up_projs  # [n_exp, dim, 2*inter]
    model.layers.{L}.moe.experts.down_projs         # [n_exp, inter, dim]
    model.layers.{L}.moe.gate.weight                # [n_exp, dim]
    model.layers.{L}.moe.gate.bias                  # [n_exp]
    model.layers.{L}.share_expert.*.weight

Note: Router gate weights and shared expert weights pass through with the same key names.
Only the expert MLP weights (gate_proj, up_proj, down_proj) need transformation.
"""

import logging
import re
from typing import Any, Optional

import torch
from torch.distributed.device_mesh import DeviceMesh

from nemo_automodel.components.checkpoint.state_dict_adapter import StateDictAdapter
from nemo_automodel.components.models.common import BackendConfig
from nemo_automodel.components.moe.config import MoEConfig
from nemo_automodel.components.moe.state_dict_utils import (
    create_dtensor_from_local,
    get_expert_range_for_rank_from_mesh,
    get_submesh,
    is_dtensor,
)

logger = logging.getLogger(__name__)


def _create_dtensor_from_local_or_reference(
    local_tensor: torch.Tensor,
    reference_dtensor: Optional["torch.Tensor"],
    device_mesh: Optional["DeviceMesh"] = None,
    rank: Optional[int] = None,
) -> torch.Tensor:
    """Create a DTensor from a local tensor.

    Prefers using reference_dtensor's mesh/placements if available (for preserving
    DTensor structure from DCP-loaded tensors). Falls back to creating a new DTensor
    using device_mesh if reference is not a DTensor.

    Args:
        local_tensor: Local portion of the tensor after transformation
        reference_dtensor: Optional DTensor to copy mesh/placements from
        device_mesh: Device mesh for EP (used if reference is not DTensor)
        rank: Current rank for device placement

    Returns:
        DTensor if mesh is available, otherwise local_tensor
    """
    from torch.distributed._tensor import DTensor

    if reference_dtensor is not None and is_dtensor(reference_dtensor):
        # Use the exact same mesh and placements from the reference
        return DTensor.from_local(local_tensor, reference_dtensor.device_mesh, reference_dtensor.placements)
    elif device_mesh is not None:
        # Create DTensor using the provided mesh
        return create_dtensor_from_local(local_tensor, device_mesh, rank)
    else:
        # No mesh available, return regular tensor
        return local_tensor


class Step3p5StateDictAdapter(StateDictAdapter):
    """Converts between HF Step3p5 checkpoints and Automodel grouped-experts native format.

    Step3p5 HF uses grouped MoELinear with shape [n_experts, out_features, in_features]:
        model.layers.{L}.moe.gate_proj.weight  # [n_exp, inter, dim]
        model.layers.{L}.moe.up_proj.weight    # [n_exp, inter, dim]
        model.layers.{L}.moe.down_proj.weight  # [n_exp, dim, inter]

    Automodel native format uses:
        model.layers.{L}.moe.experts.gate_and_up_projs  # [n_exp, dim, 2*inter]
        model.layers.{L}.moe.experts.down_projs         # [n_exp, inter, dim]
    """

    def __init__(
        self,
        config: Any,
        moe_config: MoEConfig,
        backend: BackendConfig,
        dtype: torch.dtype = torch.float32,
    ):
        self.config = config
        self.moe_config = moe_config
        self.backend = backend
        self.dtype = dtype
        self._uses_model_prefix = True

    @property
    def _hf_prefix(self) -> str:
        """Prefix for HuggingFace format keys."""
        return "model." if self._uses_model_prefix else ""

    def to_hf(
        self,
        state_dict: dict[str, Any],
        exclude_key_regex: Optional[str] = None,
        quantization: bool = False,
        **kwargs,
    ) -> dict[str, Any]:
        """Convert from native model state dict to HuggingFace format."""
        hf_state_dict = {}

        for fqn, tensor in state_dict.items():
            converted_tensors = self.convert_single_tensor_to_hf(
                fqn, tensor, exclude_key_regex=exclude_key_regex, quantization=quantization, **kwargs
            )
            for key, value in converted_tensors:
                hf_state_dict[key] = value

        return hf_state_dict

    def from_hf(
        self,
        hf_state_dict: dict[str, Any],
        device_mesh: Optional["DeviceMesh"] = None,
        **kwargs,
    ) -> dict[str, Any]:
        """Convert HF checkpoint to native format.

        Handles Step3p5's grouped MoELinear format:
        - [n_exp, inter, dim] gate_proj/up_proj -> [n_exp, dim, 2*inter] gate_and_up_projs
        - [n_exp, dim, inter] down_proj -> [n_exp, inter, dim] down_projs
        """
        # Detect prefix
        for key in hf_state_dict.keys():
            if ".moe.gate_proj.weight" in key or ".moe.up_proj.weight" in key:
                self._uses_model_prefix = key.startswith("model.")
                break

        n_experts = self.moe_config.n_routed_experts

        if device_mesh is not None:
            start_expert, end_expert = get_expert_range_for_rank_from_mesh(device_mesh, n_experts)
            rank = (
                get_submesh(device_mesh, ("ep",)).get_rank()
                if "ep" in device_mesh.mesh_dim_names
                else device_mesh.get_rank()
            )
        else:
            start_expert, end_expert = 0, n_experts
            rank = None

        state_dict: dict[str, Any] = {}
        # Track gate_proj and up_proj tensors to merge them
        pending_gate_up: dict[str, dict[str, torch.Tensor]] = {}

        # Pattern for Step3p5 grouped MoE expert weights (gate_proj, up_proj, down_proj)
        # Note: This pattern does NOT match router gate (moe.gate.weight/bias) or shared experts
        moe_pattern = re.compile(
            r"(?P<prefix>(?:model\.)?(?:language_model\.)?)layers\.(?P<layer>\d+)\.moe\.(?P<proj>gate_proj|up_proj|down_proj)\.weight"
        )

        # Pattern for router gate weights that need expert slicing for EP
        # HF Step3p5 uses moe.gate.weight and moe.router_bias (not moe.gate.bias)
        router_gate_pattern = re.compile(
            r"(?P<prefix>(?:model\.)?(?:language_model\.)?)layers\.(?P<layer>\d+)\.moe\.(?P<param>gate\.weight|gate\.bias|router_bias|gate\.e_score_correction_bias)"
        )

        for key, value in hf_state_dict.items():
            m = moe_pattern.match(key)
            router_m = router_gate_pattern.match(key)

            if m:
                prefix = m.group("prefix") or ""
                layer_num = m.group("layer")
                proj = m.group("proj")

                # For expert parallelism with regular tensors, slice the experts dimension.
                # If the value is already a DTensor (from DCP loading), it's already sharded
                # and we should not slice it again.
                if device_mesh is not None and not is_dtensor(value):
                    value = value[start_expert:end_expert]

                if proj in ("gate_proj", "up_proj"):
                    # Collect gate_proj and up_proj to merge
                    layer_key = f"{prefix}layers.{layer_num}"
                    if layer_key not in pending_gate_up:
                        pending_gate_up[layer_key] = {}

                    pending_gate_up[layer_key][proj] = value

                    # Check if we have both gate_proj and up_proj
                    if "gate_proj" in pending_gate_up[layer_key] and "up_proj" in pending_gate_up[layer_key]:
                        gate_weight = pending_gate_up[layer_key]["gate_proj"]  # [n_exp, inter, dim]
                        up_weight = pending_gate_up[layer_key]["up_proj"]  # [n_exp, inter, dim]

                        # Keep reference to original DTensor for mesh/placements (if available)
                        reference_dtensor = gate_weight if is_dtensor(gate_weight) else None

                        # Extract local tensors if DTensor for transformation
                        gate_local = gate_weight.to_local() if is_dtensor(gate_weight) else gate_weight
                        up_local = up_weight.to_local() if is_dtensor(up_weight) else up_weight

                        # Transpose: [n_exp, inter, dim] -> [n_exp, dim, inter]
                        gate_t = gate_local.transpose(1, 2)
                        up_t = up_local.transpose(1, 2)

                        # Concatenate: [n_exp, dim, 2*inter]
                        merged = torch.cat([gate_t, up_t], dim=-1).to(self.dtype)

                        native_key = f"{prefix}layers.{layer_num}.moe.experts.gate_and_up_projs"
                        # Create DTensor using reference or device_mesh
                        state_dict[native_key] = _create_dtensor_from_local_or_reference(
                            merged, reference_dtensor, device_mesh, rank
                        )

                        # Clean up
                        del pending_gate_up[layer_key]

                elif proj == "down_proj":
                    # down_proj: [n_exp, dim, inter] -> [n_exp, inter, dim]
                    # Keep reference for DTensor recreation (if available)
                    reference_dtensor = value if is_dtensor(value) else None

                    down_local = value.to_local() if is_dtensor(value) else value
                    down_t = down_local.transpose(1, 2).to(self.dtype)

                    native_key = f"{prefix}layers.{layer_num}.moe.experts.down_projs"
                    # Create DTensor using reference or device_mesh
                    state_dict[native_key] = _create_dtensor_from_local_or_reference(
                        down_t, reference_dtensor, device_mesh, rank
                    )

            elif router_m:
                # Router gate weight/bias - handle key mapping
                # Note: Router weights are NOT sliced for EP - they are replicated
                # because all ranks need full routing information
                param = router_m.group("param")
                prefix = router_m.group("prefix") or ""
                layer_num = router_m.group("layer")

                # Map HF router_bias to native gate.bias
                if param == "router_bias":
                    native_key = f"{prefix}layers.{layer_num}.moe.gate.bias"
                else:
                    native_key = key

                state_dict[native_key] = value

            else:
                # Non-MoE weights pass through unchanged (embeddings, norms, attention, shared experts, etc.)
                state_dict[key] = value

        return state_dict

    def convert_single_tensor_to_hf(
        self,
        fqn: str,
        tensor: Any,
        **kwargs,
    ) -> list[tuple[str, Any]]:
        """Convert a single tensor from native format to HuggingFace format.

        Args:
            fqn: Fully qualified name of the tensor in native format
            tensor: The tensor to convert
            **kwargs: Additional arguments for conversion

        Returns:
            List of (fqn, tensor) tuples in HuggingFace format
        """
        exclude_key_regex = kwargs.get("exclude_key_regex", None)

        result = self._convert_native_to_hf(fqn, tensor)
        if result is None:
            result = [(fqn, tensor)]

        if exclude_key_regex:
            result = [(k, v) for k, v in result if not re.match(exclude_key_regex, k)]

        return result

    def _convert_native_to_hf(
        self,
        fqn: str,
        tensor: torch.Tensor,
    ) -> list[tuple[str, torch.Tensor]] | None:
        """Convert native format expert tensors to HF Step3p5 format.

        Native: gate_and_up_projs [n_exp, dim, 2*inter] -> HF: gate_proj, up_proj [n_exp, inter, dim]
        Native: down_projs [n_exp, inter, dim] -> HF: down_proj [n_exp, dim, inter]

        Preserves DTensor structure when input is a DTensor.
        """
        from torch.distributed._tensor import DTensor

        inter_dim = self.moe_config.moe_inter_dim
        prefix = self._hf_prefix

        # Handle gate_and_up_projs
        if ".moe.experts.gate_and_up_projs" in fqn:
            layer_match = re.search(r"layers\.(\d+)", fqn)
            if layer_match is None:
                return None

            layer_num = layer_match.group(1)

            # Check if input is DTensor to preserve structure
            tensor_is_dtensor = is_dtensor(tensor)
            if tensor_is_dtensor:
                device_mesh = tensor.device_mesh
                placements = tensor.placements
                local_tensor = tensor.to_local()
            else:
                local_tensor = tensor

            # Split gate and up: [n_exp, dim, 2*inter] -> [n_exp, dim, inter] each
            gate_t = local_tensor[:, :, :inter_dim]  # [n_exp, dim, inter]
            up_t = local_tensor[:, :, inter_dim:]  # [n_exp, dim, inter]

            # Transpose: [n_exp, dim, inter] -> [n_exp, inter, dim]
            gate_weight = gate_t.transpose(1, 2).contiguous()
            up_weight = up_t.transpose(1, 2).contiguous()

            # Wrap in DTensor if original was DTensor
            if tensor_is_dtensor:
                gate_weight = DTensor.from_local(gate_weight, device_mesh, placements)
                up_weight = DTensor.from_local(up_weight, device_mesh, placements)

            return [
                (f"{prefix}layers.{layer_num}.moe.gate_proj.weight", gate_weight),
                (f"{prefix}layers.{layer_num}.moe.up_proj.weight", up_weight),
            ]

        # Handle down_projs
        if ".moe.experts.down_projs" in fqn:
            layer_match = re.search(r"layers\.(\d+)", fqn)
            if layer_match is None:
                return None

            layer_num = layer_match.group(1)

            # Check if input is DTensor to preserve structure
            tensor_is_dtensor = is_dtensor(tensor)
            if tensor_is_dtensor:
                device_mesh = tensor.device_mesh
                placements = tensor.placements
                local_tensor = tensor.to_local()
            else:
                local_tensor = tensor

            # Transpose: [n_exp, inter, dim] -> [n_exp, dim, inter]
            down_weight = local_tensor.transpose(1, 2).contiguous()

            # Wrap in DTensor if original was DTensor
            if tensor_is_dtensor:
                down_weight = DTensor.from_local(down_weight, device_mesh, placements)

            return [
                (f"{prefix}layers.{layer_num}.moe.down_proj.weight", down_weight),
            ]

        # Handle moe.gate.bias -> router_bias mapping for to_hf
        if ".moe.gate.bias" in fqn:
            layer_match = re.search(r"layers\.(\d+)", fqn)
            if layer_match is None:
                return None
            layer_num = layer_match.group(1)
            return [(f"{prefix}layers.{layer_num}.moe.router_bias", tensor)]

        # Router gate weight and shared expert weights pass through unchanged
        # These keys include: moe.gate.weight, share_expert.*
        return None



# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, Iterator, List, Optional, Union

from torch.utils.data import IterableDataset

from nemo_automodel.components.datasets.llm.column_mapped_text_instruction_dataset import (
    ColumnMappedTextInstructionDataset,
    ColumnTypes,
    _check_all_values_equal_length,
    _load_dataset,
)

logger = logging.getLogger(__name__)


def _load_streaming_dataset(
    path_or_dataset_id: Union[str, List[str]],
    split: Optional[str] = None,
    streaming: bool = False,
    name: Optional[str] = None,
    delta_storage_options: Optional[Dict[str, str]] = None,
    delta_version: Optional[int] = None,
    delta_sql_query: Optional[str] = None,
):
    """Load a dataset from HuggingFace Hub, local JSON/JSONL files, or Delta Lake tables.

    If *path_or_dataset_id* resembles a HF repo ID (i.e. of the form
    ``org/dataset`` and the path does **not** exist on the local filesystem),
    we defer to ``datasets.load_dataset`` directly. If the path is a Delta Lake
    table (prefixed with ``delta://``, ``dbfs:/``, or a directory containing
    ``_delta_log``), we load using the Delta Lake reader. Otherwise, we assume
    the argument points to one or more local JSON/JSONL files and let
    ``datasets.load_dataset`` with the *"json"* script handle the parsing.

    Args:
        path_or_dataset_id: Either a HF dataset identifier (``org/name``),
            a Delta Lake table path (``delta://path/to/table``), or
            a path / list of paths to local ``.json`` / ``.jsonl`` files.
        split: Optional split to load when retrieving a remote dataset. This
            parameter is ignored for local files and Delta Lake tables.
        streaming: Whether to stream the dataset.
        name: Optional name of the dataset configuration/subset to load
        delta_storage_options: Optional dict of storage options for Delta Lake
            cloud authentication (e.g., ``{"DATABRICKS_TOKEN": "dapi..."}``)
        delta_version: Optional specific version of the Delta table to read.
        delta_sql_query: Optional SQL query to execute against the Delta Lake source.
            This is supported when running with a SparkSession (Databricks / pyspark)
            or when using the Databricks SQL Connector. The query must return the
            columns expected by `column_mapping`.

    Returns:
        datasets.Dataset: The loaded dataset.

    Examples:
        >>> # Load from HuggingFace Hub
        >>> ds = _load_dataset("org/dataset", split="train")

        >>> # Load from local Delta Lake table
        >>> ds = _load_dataset("delta:///path/to/delta_table", streaming=True)

        >>> # Load from Databricks with authentication
        >>> ds = _load_dataset(
        ...     "delta://catalog.schema.table",
        ...     delta_storage_options={"DATABRICKS_TOKEN": "dapi..."},
        ...     streaming=True,
        ... )
    """
    # Check for Delta Lake sources first
    if isinstance(path_or_dataset_id, str):
        from nemo_automodel.components.datasets.llm.delta_lake_dataset import (
            DeltaLakeDataset,
            is_delta_lake_path,
        )

        if is_delta_lake_path(path_or_dataset_id):
            if not streaming:
                raise ValueError(
                    "Delta Lake / Databricks sources are only supported in streaming mode. "
                    "Use ColumnMappedTextInstructionIterableDataset to avoid accidental dataset materialization."
                )
            return DeltaLakeDataset(
                table_path=path_or_dataset_id,
                storage_options=delta_storage_options,
                version=delta_version,
                sql_query=delta_sql_query,
            )
    return _load_dataset(
        path_or_dataset_id,
        split=split,
        streaming=streaming,
        name=name,
    )


class ColumnMappedTextInstructionIterableDataset(IterableDataset, ColumnMappedTextInstructionDataset):
    """Streaming iterable variant that reuses the column-mapping/tokenization logic.

    This wraps a Hugging Face streaming dataset (IterableDataset from `datasets`)
    or Delta Lake table and yields tokenized samples compatible with the non-streaming
    variant, while supporting sharding and epoch-setting for deterministic shuffles upstream.

    Supports the following data sources:
    - HuggingFace Hub datasets
    - Local JSON/JSONL files
    - Delta Lake tables (via delta://, dbfs:/, or local directories with _delta_log)
    """

    def __init__(
        self,
        path_or_dataset_id: Union[str, List[str]],
        column_mapping: Dict[str, str],
        tokenizer,
        *,
        split: Optional[str] = None,
        name: Optional[str] = None,
        answer_only_loss_mask: bool = True,
        seq_length: Optional[int] = None,
        padding: Union[str, bool] = "do_not_pad",
        truncation: Union[str, bool] = "do_not_truncate",
        start_of_turn_token: Optional[str] = None,
        limit_dataset_samples: Optional[int] = None,
        repeat_on_exhaustion: bool = True,
        use_hf_chat_template: bool = False,
        delta_storage_options: Optional[Dict[str, str]] = None,
        delta_version: Optional[int] = None,
        delta_sql_query: Optional[str] = None,
    ) -> None:
        if tokenizer is None:
            raise ValueError("Tokenizer is required")
        self.tokenizer = tokenizer
        if getattr(self.tokenizer, "pad_token", None) is None:
            if hasattr(self.tokenizer, "eos_token"):
                self.tokenizer.pad_token = self.tokenizer.eos_token
            else:
                logger.warning("Setting tokenizer pad_token to ' '. tokenizer does not have `eos_token`.")
                self.tokenizer.pad_token = " "

        if ColumnTypes.Answer.value not in column_mapping:
            raise AssertionError(("Expected answer to be in column_mapping", column_mapping))
        if len(column_mapping) == 3:
            if ColumnTypes.Context.value not in column_mapping:
                raise AssertionError(("Expected context to be in column_mapping", column_mapping))
            if ColumnTypes.Question.value not in column_mapping:
                raise AssertionError(("Expected question to be in column_mapping", column_mapping))
        elif len(column_mapping) == 2:
            if ColumnTypes.Context.value not in column_mapping and ColumnTypes.Question.value not in column_mapping:
                raise AssertionError(("Expected context or question to be in column_mapping", column_mapping))
        else:
            raise ValueError(f"Expected 2 or 3 columns in column_mapping, got {len(column_mapping)}")

        self.column_mapping = column_mapping
        self.answer_only_loss_mask = answer_only_loss_mask
        self.start_of_turn_token = start_of_turn_token
        self.seq_length = seq_length
        self.padding = padding
        self.truncation = truncation
        self.use_hf_chat_template = use_hf_chat_template
        self.num_shards = getattr(self, "num_shards", 1)
        self._current_epoch_for_repeat = 0
        self.repeat_on_exhaustion = repeat_on_exhaustion
        if self.repeat_on_exhaustion is not True:
            raise ValueError("repeat_on_exhaustion must be True; False will be supported in the future.")

        # Always load in streaming mode
        ds = _load_streaming_dataset(
            path_or_dataset_id,
            split=split,
            streaming=True,
            name=name,
            delta_storage_options=delta_storage_options,
            delta_version=delta_version,
            delta_sql_query=delta_sql_query,
        )
        if limit_dataset_samples is not None and callable(getattr(ds, "take", None)):
            ds = ds.take(limit_dataset_samples)
        else:
            logger.warning("limit_dataset_samples ignored; 'take' not supported on this dataset")

        self.dataset = ds
        # Expose the underlying dataset's shard count (HF streaming uses `n_shards`).
        # This enables sharding strategies that depend on shard metadata.
        try:
            self.num_shards = int(getattr(ds, "num_shards", getattr(ds, "n_shards", self.num_shards)))
        except Exception:
            # Keep the default if the underlying dataset doesn't expose shard count.
            pass

    def __iter__(self) -> Iterator[Dict[str, List[int]]]:
        while True:
            for row in self.dataset:
                mapped = {dest: row[src] for dest, src in self.column_mapping.items() if src in row}
                # Skip rows missing required fields
                if ColumnTypes.Answer.value not in mapped:
                    continue
                tokenized = self._apply_tokenizer(mapped)  # provided by ColumnMappedTextInstructionDataset
                # Skip samples with no valid labels (aligns with non-iterable behavior)
                if not any(label != -100 for label in tokenized.get("labels", [])):
                    continue
                if not _check_all_values_equal_length(tokenized):
                    continue
                yield tokenized

            if not self.repeat_on_exhaustion:
                return
            # Wrap-around: advance epoch for deterministic reshuffle if supported and iterate again
            try:
                self._current_epoch_for_repeat += 1
                self.set_epoch(self._current_epoch_for_repeat)
            except Exception:
                pass

    def __len__(self) -> int:
        raise RuntimeError("__len__ is not supported in streaming mode.")

    def __getitem__(self, idx: int) -> Dict[str, List[int]]:
        raise RuntimeError("__getitem__ is not supported in streaming mode.")

    def set_epoch(self, epoch: int) -> None:
        if self.dataset is not None and hasattr(self.dataset, "set_epoch"):
            self.dataset.set_epoch(epoch)

    def shard(self, num_shards: int, index: int):
        if self.dataset is not None and callable(getattr(self.dataset, "shard", None)):
            self.dataset = self.dataset.shard(num_shards, index)
        return self

    def shuffle(self, buffer_size: int = 1000, seed: Optional[int] = None):
        if self.dataset is not None and callable(getattr(self.dataset, "shuffle", None)):
            self.dataset = self.dataset.shuffle(buffer_size=buffer_size, seed=seed)
        return self



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from nemo_automodel.shared.import_utils import safe_import

HAS_BNB, bitsandbytes = safe_import("bitsandbytes")
HAS_TRANSFORMERS, transformers = safe_import("transformers")


def create_bnb_config(config: dict):
    """Create BitsAndBytes config for quantization."""
    if not HAS_BNB:
        raise ImportError("bitsandbytes is required for QLora")

    if not HAS_TRANSFORMERS:
        raise ImportError("transformers is required for QLora")

    if config.load_in_4bit:
        return transformers.BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_compute_dtype=config.get("bnb_4bit_compute_dtype", "bfloat16"),
            bnb_4bit_use_double_quant=config.get("bnb_4bit_use_double_quant", True),
            bnb_4bit_quant_type=config.get("bnb_4bit_quant_type", "nf4"),
            bnb_4bit_quant_storage=config.get("bnb_4bit_quant_storage", "bfloat16"),
        )
    elif config.load_in_8bit:
        return transformers.BitsAndBytesConfig(load_in_8bit=True)
    else:
        return None


def verify_qlora_quantization(model) -> bool:
    """Verify that the model has been properly quantized."""
    for name, module in model.named_modules():
        if hasattr(module, "quant_state") and module.quant_state.__class__ == bitsandbytes.functional.QuantState:
            return True
    return False



# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import pathlib
import time

import torch
import wandb

from nemo_automodel._transformers.utils import apply_cache_compatibility_patches
from nemo_automodel.components.config._arg_parser import parse_args_and_load_config
from nemo_automodel.components.loggers.log_utils import setup_logging
from nemo_automodel.components.loggers.metric_logger import MetricsSample, build_metric_logger
from nemo_automodel.components.loggers.wandb_utils import suppress_wandb_log_messages
from nemo_automodel.components.training.rng import StatefulRNG
from nemo_automodel.components.training.utils import clip_grad_norm
from nemo_automodel.recipes._dist_setup import setup_distributed
from nemo_automodel.recipes.base_recipe import BaseRecipe
from nemo_automodel.recipes.llm.train_ft import (
    build_checkpoint_config,
    build_dataloader,
    build_distributed,
    build_lr_scheduler,
    build_model,
    build_optimizer,
    build_step_scheduler,
)

logger = logging.getLogger(__name__)


class TrainFinetuneRecipeForSequenceClassification(BaseRecipe):
    """Recipe for fine-tuning a model for sequence classification."""

    def __init__(self, cfg):
        self.cfg = cfg

    def setup(self):
        torch.cuda.reset_peak_memory_stats()
        self.dist_env = build_distributed(self.cfg.get("dist_env", {}))
        setup_logging()
        apply_cache_compatibility_patches()
        self.rng = StatefulRNG(seed=self.cfg.get("seed", 42), ranked=True)

        self.dist_setup = setup_distributed(self.cfg, world_size=self.dist_env.world_size)
        self.distributed_config = self.dist_setup.strategy_config
        self.device_mesh = self.dist_setup.device_mesh
        self.moe_mesh = self.dist_setup.moe_mesh
        self.pp_enabled = self.dist_setup.pp_enabled
        self.pipeline_config = self.dist_setup.pipeline_config

        if self.dist_env.is_main and hasattr(self.cfg, "wandb"):
            suppress_wandb_log_messages()
            # Reuse helper from NTP recipe
            from nemo_automodel.recipes.llm.train_ft import build_wandb

            run = build_wandb(self.cfg)
            logging.info("🚀 View run at {}".format(run.url))

        self._log_experiment_details()
        self._log_library_versions()

        # For classification, use standard attention implementation
        use_hf_fa2 = False

        # loss function: standard CE on logits
        self.loss_fn = torch.nn.CrossEntropyLoss()

        checkpoint_config = build_checkpoint_config(
            self.cfg.get("checkpoint", None),
            self.cfg.get("model.cache_dir", None),
            self.cfg.model.pretrained_model_name_or_path,
            True if self.cfg.get("peft", None) else False,
        )

        from nemo_automodel.components.checkpoint.checkpointing import Checkpointer

        if self.cfg.get("clip_grad_norm.max_norm", None) is not None:
            self.max_grad_norm = float(self.cfg.clip_grad_norm.max_norm)
        else:
            logging.info("No clip_grad_norm.max_norm specified in config, using default value of 1.0")
            self.max_grad_norm = 1.0

        self.checkpointer = Checkpointer(
            config=checkpoint_config,
            dp_rank=self._get_dp_rank(include_cp=True),
            tp_rank=self._get_tp_rank(),
            pp_rank=self._get_pp_rank(),
            moe_mesh=self.moe_mesh,
        )

        self.peft_config = self.cfg.instantiate_path("peft")
        model = build_model(
            cfg_model=self.cfg.model,
            cfg_peft=self.peft_config,
            seed=self.cfg.get("seed", 42),
            has_packed_sequence=use_hf_fa2,
            cfg_compile=self.cfg.get("compile", None),
            cfg_quantization=self.cfg.get("quantization", None),
            device_mesh=self.device_mesh,
            moe_mesh=self.moe_mesh,
            distributed_config=self.distributed_config,
            unfreeze_modules=["classifier"] if self.peft_config is not None else None,
        )
        self.optimizer = build_optimizer(model, self.cfg.optimizer, self.distributed_config, self.device_mesh)

        self.model_parts = [model]

        self.dataloader, self.tokenizer = build_dataloader(
            self.cfg.dataset,
            self.cfg.dataloader,
            self.cfg.model,
            cfg_ps=None,
            seed=self.cfg.get("seed", 42),
            local_batch_size=self.cfg.get("step_scheduler.local_batch_size", 1),
            global_batch_size=self.cfg.get("step_scheduler.global_batch_size", 1),
            max_steps=self.cfg.get("step_scheduler.max_steps", None),
            val_check_interval=self.cfg.get("step_scheduler.val_every_steps", None),
            dp_rank=self._get_dp_rank(),
            dp_world_size=self._get_dp_group_size(),
            pp_enabled=False,
            supports_seq_lens=False,
        )

        self.val_dataloader = None
        if "validation_dataset" in self.cfg:
            self.val_dataloader, _ = build_dataloader(
                self.cfg.validation_dataset,
                self.cfg.validation_dataloader,
                self.cfg.model,
                cfg_ps=None,
                seed=self.cfg.get("seed", 42),
                local_batch_size=self.cfg.get("step_scheduler.local_batch_size", 1),
                global_batch_size=self.cfg.get("step_scheduler.global_batch_size", 1),
                max_steps=self.cfg.get("step_scheduler.max_steps", None),
                val_check_interval=self.cfg.get("step_scheduler.val_every_steps", None),
                dp_rank=self._get_dp_rank(),
                dp_world_size=self._get_dp_group_size(),
                pp_enabled=False,
            )

        self.best_metric_key = self.cfg.get("checkpoint.best_metric_key", "default")
        self.step_scheduler = build_step_scheduler(
            self.cfg.get("step_scheduler", None),
            self.dataloader,
            self._get_dp_group_size(),
            local_batch_size=self.cfg.get("step_scheduler.local_batch_size", 1),
        )

        self.lr_scheduler = build_lr_scheduler(self.cfg.get("lr_scheduler", None), self.optimizer, self.step_scheduler)

        self._log_model_and_optimizer_details(self.model_parts, self.optimizer, self.lr_scheduler)

        restore_from = self.cfg.get("checkpoint.restore_from", None)
        self.metric_logger_train = build_metric_logger(
            pathlib.Path(self.checkpointer.config.checkpoint_dir) / "training.jsonl"
        )
        self.metric_logger_valid = build_metric_logger(
            pathlib.Path(self.checkpointer.config.checkpoint_dir) / "validation.jsonl"
        )
        self.load_checkpoint(restore_from)
        self._log_step_scheduler_details(self.step_scheduler)

    def run_train_validation_loop(self):
        for mp in self.model_parts:
            mp.train()
        self.timestamp = time.perf_counter()

        for epoch in self.step_scheduler.epochs:
            self.step_scheduler.set_epoch(epoch)
            for batches in self.step_scheduler:
                train_log_data = self._run_train_optim_step(batches)
                self.log_train_metrics(train_log_data)

                val_loss = {}
                if self.step_scheduler.is_val_step and self.val_dataloader is not None:
                    val_log_data = self._validate_one_epoch(self.val_dataloader)
                    val_loss["val_loss"] = val_log_data.metrics["val_loss"]
                    self.log_val_metrics(val_log_data)
                    for mp in self.model_parts:
                        mp.train()

                if self.step_scheduler.is_ckpt_step:
                    self.save_checkpoint(
                        epoch,
                        self.step_scheduler.step,
                        train_log_data.metrics["loss"],
                        val_loss,
                        best_metric_key=self.best_metric_key,
                    )

        self.metric_logger_train.close()
        self.metric_logger_valid.close()
        self.checkpointer.close()

    def _run_train_optim_step(self, batches):
        model = self.model_parts[0]
        losses = []
        all_preds = []
        all_labels = []

        # Count input tokens for throughput calculation (excluding padding)
        num_tokens_in_batch = torch.tensor(
            sum(batch["attention_mask"].sum().item() for batch in batches),
            dtype=torch.long,
        )
        num_tokens_in_batch = self._dp_allreduce(num_tokens_in_batch).item()

        for batch in batches:
            batch = {
                k: (v.to(self.dist_env.device, non_blocking=True) if v is not None else None) for k, v in batch.items()
            }
            labels = batch.pop("labels")
            out = model(**batch)
            logits = getattr(out, "logits", out)
            loss = self.loss_fn(logits, labels.view(-1))
            losses.append(loss.detach().clone())

            # Collect predictions for accuracy calculation
            preds = torch.argmax(logits, dim=-1)
            all_preds.append(preds.detach())
            all_labels.append(labels.view(-1).detach())
            (loss * self._get_dp_group_size(include_cp=True)).backward()

        # Calculate gradient norm (distributed-aware)
        grad_norm = clip_grad_norm(
            max_grad_norm=self.max_grad_norm,
            model_parts=self.model_parts,
            norm_type=2.0,
            pp_enabled=self._get_pp_rank() != 0 if hasattr(self, "_get_pp_rank") else False,
            device_mesh=self.device_mesh,
        )

        # Calculate accuracy
        all_preds = torch.cat(all_preds)
        all_labels = torch.cat(all_labels)
        correct = (all_preds == all_labels).sum()
        total = all_labels.numel()
        accuracy = correct.float() / total

        # Sync accuracy across distributed ranks if needed
        if self._get_dp_group_size(include_cp=True) > 1:
            correct = self._dp_allreduce(correct.float(), include_cp=True)
            total_across_ranks = self._dp_allreduce(
                torch.tensor(total, device=correct.device, dtype=torch.float), include_cp=True
            )
            accuracy = correct / total_across_ranks

        for opt in self.optimizer:
            opt.step()
            opt.zero_grad()
        if self.lr_scheduler is not None:
            for scheduler in self.lr_scheduler:
                scheduler.step(1)

        # Calculate throughput (tokens per second)
        t = time.perf_counter()
        time_delta = t - self.timestamp
        self.timestamp = t
        tps = num_tokens_in_batch / time_delta

        total_loss = torch.sum(torch.stack(losses))
        total_loss = self._dp_allreduce(total_loss, include_cp=True).detach()
        loss = total_loss / len(batches)

        return MetricsSample(
            step=self.step_scheduler.step,
            epoch=self.step_scheduler.epoch,
            metrics={
                "loss": loss,
                "accuracy": accuracy.item() if isinstance(accuracy, torch.Tensor) else accuracy,
                "grad_norm": grad_norm,
                "lr": self.optimizer[0].param_groups[0]["lr"],
                "mem": torch.cuda.max_memory_allocated() / 1024**3,
                "tps": tps,
                "tps_per_gpu": tps / self._get_cp_group_size() / max(self._get_dp_group_size(), 1),
            },
        )

    @torch.no_grad()
    def _validate_one_epoch(self, dataloader):
        model = self.model_parts[0]
        model.eval()
        total_loss = 0.0
        all_preds = []
        all_labels = []
        count = 0

        for batch in dataloader:
            batch = {
                k: (v.to(self.dist_env.device, non_blocking=True) if v is not None else None) for k, v in batch.items()
            }
            labels = batch.pop("labels")
            out = model(**batch)
            logits = getattr(out, "logits", out)
            loss = self.loss_fn(logits, labels.view(-1))
            total_loss += loss.detach()

            # Collect predictions for accuracy
            preds = torch.argmax(logits, dim=-1)
            all_preds.append(preds)
            all_labels.append(labels.view(-1))
            count += 1

        total_loss = total_loss if count == 0 else total_loss / count

        # Calculate accuracy
        if len(all_preds) > 0:
            all_preds = torch.cat(all_preds)
            all_labels = torch.cat(all_labels)
            correct = (all_preds == all_labels).sum()
            total = all_labels.numel()
            accuracy = correct.float() / total

            # Sync across distributed ranks if needed
            if self._get_dp_group_size(include_cp=True) > 1:
                correct = self._dp_allreduce(correct.float(), include_cp=True)
                total_across_ranks = self._dp_allreduce(
                    torch.tensor(total, device=correct.device, dtype=torch.float), include_cp=True
                )
                accuracy = correct / total_across_ranks
        else:
            accuracy = 0.0

        return MetricsSample(
            step=self.step_scheduler.step,
            epoch=self.step_scheduler.epoch,
            metrics={
                "val_loss": total_loss,
                "val_accuracy": accuracy.item() if isinstance(accuracy, torch.Tensor) else accuracy,
                "lr": self.optimizer[0].param_groups[0]["lr"],
                "mem": torch.cuda.max_memory_allocated() / 1024**3,
            },
        )

    def log_val_metrics(self, log_data):
        """Log metrics to wandb and other loggers
        Args:
            log_data: MetricsSample object, containing:
                step: int, the current step.
                epoch: int, the current epoch.
                metrics: Dict[str, float], containing:
                    "val_loss": Validation loss.
                    "lr": Learning rate.
                    "num_label_tokens": Number of label tokens.
                    "mem": Memory allocated.
        """

        # Pipeline parallelism does not support validation -> log_data is None
        if not self.dist_env.is_main or log_data is None:
            return

        if wandb.run is not None:
            wandb.log(log_data.to_dict(), step=log_data.step)

        # JSONL validation log
        self.metric_logger_valid.log(log_data)

        logging.info(
            "[val] step {} | epoch {} | loss {:.4f} | accuracy {:.4f} | lr {:.2e}".format(
                log_data.step,
                log_data.epoch,
                log_data.metrics["val_loss"],
                log_data.metrics["val_accuracy"],
                log_data.metrics["lr"],
            )
        )

    def log_train_metrics(self, log_data):
        """Log metrics to wandb and other loggers.

        Args:
            log_data: MetricsSample object, containing:
                step: int, the current step.
                epoch: int, the current epoch.
                metrics: Dict[str, float], containing:
                    "loss": Training loss.
                    "accuracy": Training accuracy.
                    "grad_norm": Gradient norm from the training step.
                    "lr": Learning rate.
                    "mem": Memory allocated.
                    "tps": Tokens per second (throughput).
                    "tps_per_gpu": Tokens per second per GPU.
        """
        if not self.dist_env.is_main:
            return

        # Log to remote services (WandB) according to step_scheduler frequency
        if self.step_scheduler.is_remote_logging_step:
            if wandb.run is not None:
                wandb.log(log_data.to_dict(), step=self.step_scheduler.step)

        # JSONL training log (always log for detailed local records)
        self.metric_logger_train.log(log_data)
        logging.info(
            "step {} | epoch {} | loss {:.4f} | accuracy {:.4f} | grad_norm {:.4f} | lr {:.2e} | mem {:.2f} GiB | tps {:.2f}({:.2f}/gpu)".format(
                log_data.step,
                log_data.epoch,
                log_data.metrics["loss"],
                log_data.metrics["accuracy"],
                log_data.metrics["grad_norm"],
                log_data.metrics["lr"],
                log_data.metrics["mem"],
                log_data.metrics["tps"],
                log_data.metrics["tps_per_gpu"],
            )
        )
        torch.cuda.reset_peak_memory_stats()


def main(config_path: str | None = None):
    if config_path is None:
        config_path = (
            pathlib.Path(__file__).parent.resolve()
            / "../.."
            / "examples/llm_sequence_classification/yelp/yelp_bert.yaml"
        )
    cfg = parse_args_and_load_config(config_path)
    trainer = TrainFinetuneRecipeForSequenceClassification(cfg)
    trainer.setup()
    trainer.run_train_validation_loop()


if __name__ == "__main__":
    main()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DeepSeek V3.2 Layers.

Contains the DeepseekV32Indexer for top-k sparse attention selection
and DeepseekV32MLA which integrates the indexer with Multi-head Latent Attention.
"""

from typing import Any

import torch
from torch import nn

# Try to import fast_hadamard_transform, fall back to torch implementation
try:
    from fast_hadamard_transform import hadamard_transform

    _FAST_HADAMARD_AVAILABLE = True
except ImportError:
    _FAST_HADAMARD_AVAILABLE = False

    # Taken from https://github.com/HazyResearch/structured-nets/blob/master/pytorch/structure/hadamard.py#L26
    def hadamard_transform_torch(u, scale: float, normalize=False):
        """Multiply H_n @ u where H_n is the Hadamard matrix of dimension n x n.
        n must be a power of 2.
        Parameters:
            u: Tensor of shape (..., n)
            normalize: if True, divide the result by 2^{m/2} where m = log_2(n).
        Returns:
            product: Tensor of shape (..., n)
        """
        import math

        n = u.shape[-1]
        m = int(math.log2(n))
        assert n == 1 << m, "n must be a power of 2"
        x = u.unsqueeze(-1)
        for _ in range(m):
            x = torch.cat((x[..., ::2, :] + x[..., 1::2, :], x[..., ::2, :] - x[..., 1::2, :]), dim=-1)
        x = x.squeeze(-2) / 2 ** (m / 2) if normalize else x.squeeze(-2)
        return x * scale

    def hadamard_transform(x: torch.Tensor, scale: float) -> torch.Tensor:
        """Fallback hadamard_transform when fast_hadamard_transform is not available."""
        return hadamard_transform_torch(x, scale)


from nemo_automodel.components.attention.utils import (
    initialize_attn_module_and_func,
    postprocess_output_for_attn,
    preprocess_args_and_kwargs_for_attn,
)
from nemo_automodel.components.models.common import (
    BackendConfig,
    initialize_linear_module,
    initialize_rms_norm_module,
)
from nemo_automodel.components.models.deepseek_v3.rope_utils import (
    apply_rotary_emb,
    yarn_get_mscale,
)
from nemo_automodel.components.models.deepseek_v32.config import DeepseekV32Config


def _rotate_activation(x: torch.Tensor) -> torch.Tensor:
    """Apply Hadamard rotation activation.

    Reference:
        https://github.com/deepseek-ai/DeepSeek-V3.2-Exp/blob/main/inference/model.py#L424-L428

    Args:
        x: Input tensor (must be bfloat16).

    Returns:
        Rotated tensor.
    """
    if x.dtype != torch.bfloat16:
        x = x.to(torch.bfloat16)
    hidden_size = x.size(-1)
    return hadamard_transform(x, scale=hidden_size**-0.5)


class DeepseekV32Indexer(nn.Module):
    """Indexer for top-k sparse attention selection.

    Based on the official DeepSeek V3.2 training implementation. Computes attention
    scores between queries and keys with per-head weights, applies ReLU activation,
    then selects the top-k positions to attend to.

    Key features:
    - Uses LayerNorm (not RMSNorm) for key normalization
    - Has a weights_proj that learns per-head importance weights
    - Optional Hadamard transform (rotate_activation) on Q and K
    - ReLU activation on attention scores before weighting
    """

    def __init__(self, config: DeepseekV32Config, backend: BackendConfig):
        super().__init__()

        self.num_heads = config.index_n_heads
        self.head_dim = config.index_head_dim
        self.qk_rope_head_dim = config.qk_rope_head_dim
        self.qk_nope_head_dim = self.head_dim - self.qk_rope_head_dim
        self.index_topk = config.index_topk
        self.q_lora_rank = config.q_lora_rank
        self.hidden_size = config.hidden_size
        self.softmax_scale = self.head_dim**-0.5

        self.backend = backend
        linear_impl = backend.linear

        # Project Q from q_lora residual -> num_heads * head_dim
        self.wq_b = initialize_linear_module(
            linear_impl=linear_impl,
            in_features=self.q_lora_rank,
            out_features=self.num_heads * self.head_dim,
            bias=False,
        )

        # Project K from hidden states -> single head_dim (shared across heads)
        self.wk = initialize_linear_module(
            linear_impl=linear_impl,
            in_features=self.hidden_size,
            out_features=self.head_dim,
            bias=False,
        )

        # LayerNorm for K (official uses LayerNorm, not RMSNorm)
        self.k_norm = nn.LayerNorm(self.head_dim)

        # Per-head weight projection from hidden states
        self.weights_proj = initialize_linear_module(
            linear_impl=linear_impl,
            in_features=self.hidden_size,
            out_features=self.num_heads,
            bias=False,
        )

    def forward(
        self,
        x: torch.Tensor,
        q_resid: torch.Tensor,
        freqs_cis: torch.Tensor,
        attention_mask: torch.Tensor | None = None,
        **attn_kwargs: Any,
    ) -> torch.Tensor:
        """Compute top-k indices for sparse attention.

        Args:
            x: Hidden states [B, S, hidden] or [T, hidden] for thd format
            q_resid: Q lora residual from MLA [B, S, q_lora_rank] or [T, q_lora_rank]
            freqs_cis: RoPE frequencies
            attention_mask: Optional attention mask
            **attn_kwargs: Additional attention kwargs (cu_seqlens, etc.)

        Returns:
            topk_indices: Indices of top-k positions [B, S, topk] or [T, topk]
        """
        if len(x.shape) == 2:
            qkv_format = "thd"
            num_tokens = x.shape[0]
            bsz = 1
            seq_len = num_tokens
        else:
            qkv_format = "bshd"
            bsz, seq_len, _ = x.size()

        # Project Q from q_lora residual
        q = self.wq_b(q_resid)
        if qkv_format == "thd":
            q = q.view(num_tokens, self.num_heads, self.head_dim)
        else:
            q = q.view(bsz, seq_len, self.num_heads, self.head_dim)

        # Split Q into nope and pe parts (nope first, then pe - matching training code)
        q_nope, q_pe = torch.split(q, [self.qk_nope_head_dim, self.qk_rope_head_dim], dim=-1)

        # Project K from hidden states
        k = self.k_norm(self.wk(x))

        # Split K into nope and pe parts
        k_nope, k_pe = torch.split(k, [self.qk_nope_head_dim, self.qk_rope_head_dim], dim=-1)

        # Apply RoPE to the pe parts
        head_unsqueeze_dim = 2 if qkv_format == "bshd" else 1
        q_pe = apply_rotary_emb(q_pe, freqs_cis, qkv_format=qkv_format)
        k_pe = apply_rotary_emb(k_pe, freqs_cis, qkv_format=qkv_format, unsqueeze_dim=head_unsqueeze_dim)
        k_pe = k_pe.squeeze(head_unsqueeze_dim)

        # Combine nope and pe parts (nope first, matching training code)
        q = torch.cat([q_nope, q_pe], dim=-1)
        k = torch.cat([k_nope, k_pe], dim=-1)

        # Apply optional Hadamard rotation (if fast_hadamard_transform is available)
        q = _rotate_activation(q)
        k = _rotate_activation(k)

        # Compute per-head weights from hidden states
        # weights: [B, S, H] or [T, H]
        # Scale: (n_heads ** -0.5) * softmax_scale
        weights = self.weights_proj(x).float() * (self.num_heads**-0.5) * self.softmax_scale

        # Expand K to all heads
        if qkv_format == "thd":
            k = k.unsqueeze(1).expand(num_tokens, self.num_heads, self.head_dim)
        else:
            k = k.unsqueeze(2).expand(bsz, seq_len, self.num_heads, self.head_dim)

        # Compute attention scores: Q @ K^T with ReLU activation
        if qkv_format == "thd":
            # [T, H, D] -> [H, T, D]
            q = q.transpose(0, 1)
            k = k.transpose(0, 1)
            # scores: [H, T, T]
            scores = torch.bmm(q.float(), k.float().transpose(-2, -1))
            # Apply ReLU activation (per training implementation)
            scores = torch.relu(scores)
            # Apply per-head weights: [T, H] -> [H, T, 1]
            weights = weights.transpose(0, 1).unsqueeze(-1)
            scores = scores * weights  # [H, T, T]
            # Sum over heads
            scores = scores.sum(dim=0)  # [T, T]
        else:
            # [B, S, H, D] -> [B, H, S, D]
            q = q.transpose(1, 2)
            k = k.transpose(1, 2)
            # scores: [B, H, S, S]
            scores = torch.matmul(q.float(), k.float().transpose(-2, -1))
            # Apply ReLU activation (per training implementation)
            scores = torch.relu(scores)
            # Apply per-head weights: [B, S, H] -> [B, H, S, 1]
            weights = weights.transpose(1, 2).unsqueeze(-1)
            scores = scores * weights  # [B, H, S, S]
            # Sum over heads
            scores = scores.sum(dim=1)  # [B, S, S]

        # Apply attention mask if provided
        if attention_mask is not None:
            if qkv_format == "bshd":
                scores = scores + attention_mask.squeeze(1)
            else:
                if attention_mask.dim() == 4:
                    scores = scores + attention_mask.squeeze(0).squeeze(0)
                else:
                    scores = scores + attention_mask

        # Select top-k indices
        actual_topk = min(self.index_topk, seq_len)
        topk_indices = scores.topk(actual_topk, dim=-1).indices

        return topk_indices

    def init_weights(self, init_std: float = 0.02):
        for module in [self.wq_b, self.wk, self.weights_proj]:
            if hasattr(module, "weight"):
                nn.init.trunc_normal_(module.weight, mean=0.0, std=init_std)
        self.k_norm.reset_parameters()


class DeepseekV32MLA(nn.Module):
    """Multi-head Latent Attention with Indexer for sparse attention.

    This extends the V3 MLA with an Indexer module that performs
    top-k selection for sparse attention. The indexer uses the
    q_lora residual and hidden states to compute which positions
    to attend to.
    """

    def __init__(self, config: DeepseekV32Config, backend: BackendConfig):
        super().__init__()

        self.n_heads = config.num_attention_heads
        self.q_lora_rank = config.q_lora_rank
        self.kv_lora_rank = config.kv_lora_rank
        self.qk_nope_head_dim = config.qk_nope_head_dim
        self.qk_rope_head_dim = config.qk_rope_head_dim
        self.qk_head_dim = (
            config.qk_head_dim if hasattr(config, "qk_head_dim") else (self.qk_nope_head_dim + self.qk_rope_head_dim)
        )
        self.v_head_dim = config.v_head_dim
        self.index_topk = config.index_topk

        self.backend = backend
        self.rope_fusion = backend.rope_fusion
        attn_impl = backend.attn
        linear_impl = backend.linear
        rms_norm_impl = backend.rms_norm

        hidden_size = config.hidden_size

        # V3.2 always uses q_lora (q_lora_rank is not None)
        self.q_a_proj = initialize_linear_module(
            linear_impl=linear_impl, in_features=hidden_size, out_features=self.q_lora_rank, bias=False
        )
        self.q_a_layernorm = initialize_rms_norm_module(rms_norm_impl=rms_norm_impl, dim=self.q_lora_rank)
        self.q_b_proj = initialize_linear_module(
            linear_impl=linear_impl,
            in_features=self.q_lora_rank,
            out_features=self.n_heads * self.qk_head_dim,
            bias=False,
        )

        self.kv_a_proj_with_mqa = initialize_linear_module(
            linear_impl=linear_impl,
            in_features=hidden_size,
            out_features=self.kv_lora_rank + self.qk_rope_head_dim,
            bias=False,
        )
        self.kv_a_layernorm = initialize_rms_norm_module(rms_norm_impl=rms_norm_impl, dim=self.kv_lora_rank)
        self.kv_b_proj = initialize_linear_module(
            linear_impl=linear_impl,
            in_features=self.kv_lora_rank,
            out_features=self.n_heads * (self.qk_nope_head_dim + self.v_head_dim),
            bias=False,
        )
        self.o_proj = initialize_linear_module(
            linear_impl=linear_impl,
            in_features=self.n_heads * self.v_head_dim,
            out_features=hidden_size,
            bias=False,
        )
        self.softmax_scale = self.qk_head_dim**-0.5

        rope_parameters = config.rope_parameters if hasattr(config, "rope_parameters") else config.rope_scaling
        if rope_parameters and all(
            map(lambda x: x in rope_parameters, ["factor", "mscale", "original_max_position_embeddings"])
        ):
            factor = rope_parameters["factor"]
            mscale = rope_parameters["mscale"]
            original_seq_len = rope_parameters["original_max_position_embeddings"]
            if config.max_position_embeddings > original_seq_len:
                mscale = yarn_get_mscale(factor, mscale)
            self.softmax_scale = self.softmax_scale * mscale * mscale

        self.attn_module, self.attn_func = initialize_attn_module_and_func(
            attn_impl=attn_impl,
            num_attention_heads=self.n_heads,
            num_qk_channels=self.qk_head_dim,
            num_v_channels=self.v_head_dim,
            softmax_scale=self.softmax_scale,
        )

        # Initialize the Indexer
        self.indexer = DeepseekV32Indexer(config, backend)

    def _build_sparse_mask(
        self,
        topk_indices: torch.Tensor,
        seq_len: int,
        qkv_format: str,
        bsz: int = 1,
        n_heads: int = 1,
        dtype: torch.dtype = torch.bfloat16,
        attention_mask: torch.Tensor | None = None,
        union_across_batches: bool = False,
    ) -> torch.Tensor:
        """Build a sparse attention mask/bias from top-k indices.

        Creates a mask tensor where non-top-k positions are set to -inf.
        Works for both TE (core_attention_bias) and SDPA (attn_mask).

        Uses the same efficient pattern as the official DeepSeek inference code:
        `torch.full(..., -inf).scatter_(-1, topk_indices, 0)`

        Args:
            topk_indices: Indices of top-k positions [B, S, topk] or [T, topk]
            seq_len: Sequence length
            qkv_format: 'bshd' or 'thd'
            bsz: Batch size (only used for bshd format)
            n_heads: Number of attention heads to expand to
            dtype: Data type for the output tensor
            attention_mask: Optional attention mask to combine with (for SDPA)
            union_across_batches: If True, union top-k across batches (for TE);
                                  if False, keep per-batch masks (for SDPA)

        Returns:
            sparse_mask: Mask tensor with shape:
                - [1, n_heads, S, S] if union_across_batches=True
                - [B, n_heads, S, S] if union_across_batches=False (bshd)
                - [1, n_heads, T, T] for thd format
        """
        device = topk_indices.device

        if qkv_format == "thd":
            num_tokens = topk_indices.shape[0]
            # Create mask directly in final shape [1, n_heads, T, T]
            # All heads share the same mask, so we create [T, T] and expand
            sparse_mask = torch.full((num_tokens, num_tokens), float("-inf"), device=device, dtype=dtype).scatter_(
                -1, topk_indices, 0.0
            )
            # expand creates a view, contiguous makes a copy
            sparse_mask = sparse_mask.view(1, 1, num_tokens, num_tokens).expand(1, n_heads, -1, -1).contiguous()
        else:
            if union_across_batches:
                # For TE: create [B, S, S], scatter, union via max, then expand
                sparse_mask = torch.full((bsz, seq_len, seq_len), float("-inf"), device=device, dtype=dtype).scatter_(
                    -1, topk_indices, 0.0
                )
                # Union: max(0, -inf) = 0 for any position selected in any batch
                sparse_mask = sparse_mask.max(dim=0, keepdim=True).values
                sparse_mask = sparse_mask.view(1, 1, seq_len, seq_len).expand(1, n_heads, -1, -1).contiguous()
            else:
                # For SDPA: create [B, S, S], scatter, expand (no contiguous needed)
                sparse_mask = torch.full((bsz, seq_len, seq_len), float("-inf"), device=device, dtype=dtype).scatter_(
                    -1, topk_indices, 0.0
                )
                sparse_mask = sparse_mask.unsqueeze(1).expand(-1, n_heads, -1, -1)

        # Combine with existing attention mask if provided
        if attention_mask is not None:
            sparse_mask = attention_mask + sparse_mask

        return sparse_mask

    def forward(
        self,
        x: torch.Tensor,
        freqs_cis: torch.Tensor,
        attention_mask: torch.Tensor | None = None,
        **attn_kwargs: Any,
    ):
        if len(x.shape) == 2:
            qkv_format = "thd"
            num_tokens = x.shape[0]
            bsz = 1
            seq_len = num_tokens
        else:
            qkv_format = "bshd"
            bsz, seq_len, _ = x.size()

        # Compute q_resid for indexer and main attention path
        q_resid = self.q_a_layernorm(self.q_a_proj(x))

        # Get top-k indices from indexer
        topk_indices = self.indexer(x, q_resid, freqs_cis, attention_mask, **attn_kwargs)

        # Build sparse bias/mask from top-k indices based on backend
        if self.backend.attn == "te":
            # For TE: build sparse bias for core_attention_bias (must match Q/K/V dtype)
            # Union across batches since TE expects [1, n_heads, S, S]
            sparse_mask = self._build_sparse_mask(
                topk_indices,
                seq_len,
                qkv_format,
                bsz,
                n_heads=self.n_heads,
                dtype=x.dtype,
                attention_mask=None,
                union_across_batches=True,
            )
        else:
            # For SDPA: build sparse mask, keep per-batch masks
            sparse_mask = self._build_sparse_mask(
                topk_indices,
                seq_len,
                qkv_format,
                bsz,
                n_heads=1,
                dtype=torch.float32,
                attention_mask=attention_mask,
                union_across_batches=False,
            )

        # Compute Q from q_resid
        q = self.q_b_proj(q_resid)

        if qkv_format == "thd":
            q = q.view(num_tokens, self.n_heads, self.qk_head_dim)
        else:
            q = q.view(bsz, seq_len, self.n_heads, self.qk_head_dim)

        q_nope, q_pe = torch.split(q, [self.qk_nope_head_dim, self.qk_rope_head_dim], dim=-1)

        kv = self.kv_a_proj_with_mqa(x)
        kv, k_pe = torch.split(kv, [self.kv_lora_rank, self.qk_rope_head_dim], dim=-1)
        kv = self.kv_a_layernorm(kv)

        # For MLA, k_pe needs an extra head dimension for apply_rotary_emb
        head_unsqueeze_dim = 2 if qkv_format == "bshd" else 1
        k_pe = k_pe.unsqueeze(head_unsqueeze_dim)

        # Apply rotary embeddings to q_pe and k_pe
        q_pe = apply_rotary_emb(q_pe, freqs_cis, qkv_format=qkv_format)
        k_pe = apply_rotary_emb(k_pe, freqs_cis, qkv_format=qkv_format)

        # Remove the head dimension we added to k_pe
        k_pe = k_pe.squeeze(head_unsqueeze_dim)

        q = torch.cat([q_nope, q_pe], dim=-1)

        kv = self.kv_b_proj(kv)
        if qkv_format == "thd":
            kv = kv.view(num_tokens, self.n_heads, self.qk_nope_head_dim + self.v_head_dim)
            k_pe = k_pe.unsqueeze(1).expand([num_tokens, self.n_heads, self.qk_rope_head_dim])
        else:
            kv = kv.view(bsz, seq_len, self.n_heads, self.qk_nope_head_dim + self.v_head_dim)
            k_pe = k_pe.unsqueeze(2).expand([bsz, seq_len, self.n_heads, self.qk_rope_head_dim])

        k_nope, v = torch.split(kv, [self.qk_nope_head_dim, self.v_head_dim], dim=-1)
        k = torch.cat([k_nope, k_pe], dim=-1)

        # Handle attention based on backend
        if self.backend.attn == "te":
            # For TE: use core_attention_bias for sparse attention
            q, k, v, _attn_kwargs = preprocess_args_and_kwargs_for_attn(
                q, k, v, attention_mask, self.backend.attn, **attn_kwargs
            )
            # Add sparse mask as core_attention_bias
            _attn_kwargs["core_attention_bias_type"] = "post_scale_bias"
            _attn_kwargs["core_attention_bias"] = sparse_mask
        else:
            # For SDPA: use sparse mask (already combined with attention_mask)
            q, k, v, _attn_kwargs = preprocess_args_and_kwargs_for_attn(
                q, k, v, sparse_mask, self.backend.attn, **attn_kwargs
            )

        x = self.attn_func(q, k, v, **_attn_kwargs)
        x = postprocess_output_for_attn(x, self.backend.attn)

        flatten_dim = 2 if qkv_format == "bshd" else 1
        x = self.o_proj(x.flatten(flatten_dim))
        return x

    def init_weights(self, _buffer_device: torch.device, init_std: float = 0.02):
        linear_list = [
            self.q_a_proj,
            self.q_b_proj,
            self.kv_a_proj_with_mqa,
            self.kv_b_proj,
            self.o_proj,
        ]

        for linear in linear_list:
            nn.init.trunc_normal_(linear.weight, mean=0.0, std=init_std)

        norms = [self.kv_a_layernorm, self.q_a_layernorm]
        for norm in norms:
            norm.reset_parameters()

        # Initialize indexer weights
        self.indexer.init_weights(init_std)



# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Knowledge Distillation recipe for next-token prediction with NeMo-AutoModel.

This recipe fine-tunes a *student* model using the logits of a frozen *teacher* model. It
extends ``FinetuneRecipeForNextTokenPrediction`` adding:

1. teacher_model – an additional HF/NeMo model loaded in ``eval`` mode
2. kd_loss_fn     – KL-divergence between temperature-scaled distributions
3. kd_ratio       – linear mix between CE loss and KD loss

The training loop is copied from the parent class but the loss becomes:
    loss = (1-kd_ratio) * ce_loss + kd_ratio * kd_loss

The file exposes ``KnowledgeDistillationRecipeForNextTokenPrediction`` and a
``main`` entry-point so it can be launched exactly the same way as other
recipes:

    python -m torch.distributed.run --nproc-per-node=8 \
        nemo_automodel/recipes/llm/knowledge_distillation.py \
        -c examples/llm/llama_3_2_1b_kd.yaml
"""

from __future__ import annotations

import logging
import time
from contextlib import nullcontext
from typing import Any, Dict, Optional

import torch
import wandb
from torchao.float8 import precompute_float8_dynamic_scale_for_fsdp

from nemo_automodel._transformers.auto_tokenizer import NeMoAutoTokenizer
from nemo_automodel.components.config._arg_parser import parse_args_and_load_config
from nemo_automodel.components.distributed.cp_utils import make_cp_batch_and_ctx
from nemo_automodel.components.distributed.utils import get_sync_ctx
from nemo_automodel.components.loggers.metric_logger import MetricsSample
from nemo_automodel.components.loss.linear_ce import FusedLinearCrossEntropy
from nemo_automodel.components.training.rng import ScopedRNG
from nemo_automodel.components.training.utils import ScopedModuleOffloading, count_tail_padding
from nemo_automodel.recipes.llm.train_ft import (
    TrainFinetuneRecipeForNextTokenPrediction,
    calculate_loss,
)

logger = logging.getLogger(__name__)


def _build_kd_loss_fn(cfg_kd):
    if cfg_kd is None:
        logger.info("No KD loss function provided, using KLDivLoss")
        return torch.nn.KLDivLoss(reduction="batchmean")
    return cfg_kd.instantiate()


def _build_teacher_model(
    cfg_teacher,
    seed,
    has_packed_sequence,
    device_mesh=None,
    moe_mesh=None,
    distributed_config=None,
    device=None,
):
    """Build and initialize the teacher model for knowledge distillation.

    Uses the same infrastructure as student model (NeMoAutoModelForCausalLM) but without
    PEFT, FP8, or QAT since the teacher should be frozen in full precision.

    Args:
        cfg_teacher: Configuration for teacher model instantiation.
        seed: Random seed for reproducibility.
        has_packed_sequence: Whether using packed sequences.
        device_mesh: Device mesh for distributed training.
        moe_mesh: MOE mesh for expert parallelism.
        distributed_config: Strategy-specific distributed config.
        device: Device to place the teacher model on.

    Returns:
        The frozen teacher model ready for inference.

    Note:
        The `offload_teacher_model` config option is not supported with this approach.
        Device placement is handled internally by NeMoAutoModelForCausalLM infrastructure.
    """

    assert cfg_teacher is not None, "`teacher_model` section missing from YAML config"
    logger.info("Instantiating teacher model")

    # Build teacher model using the same infrastructure as student
    # but without PEFT/FP8/QAT (teacher should be frozen in full precision)
    with ScopedRNG(seed=seed, ranked=True):
        kwargs: Dict[str, Any] = {
            "has_packed_sequence": has_packed_sequence,
            "device_mesh": device_mesh,
            "moe_mesh": moe_mesh,
            "distributed_config": distributed_config,
        }

        teacher_model = cfg_teacher.instantiate(**kwargs)

        # Ensure the teacher model is on the correct device
        teacher_model = teacher_model.to(device)

        # Set teacher to eval mode and freeze parameters
        teacher_model.eval()
        for p in teacher_model.parameters():
            p.requires_grad_(False)

        return teacher_model


def _verify_tokenizer_compatibility(student_cfg, teacher_cfg, trust_remote_code=True):
    if student_cfg is None or teacher_cfg is None:
        raise ValueError("Student and teacher model configs are required")
    student_tokenizer = NeMoAutoTokenizer.from_pretrained(
        student_cfg.pretrained_model_name_or_path, trust_remote_code=trust_remote_code
    )
    teacher_tokenizer = NeMoAutoTokenizer.from_pretrained(
        teacher_cfg.pretrained_model_name_or_path, trust_remote_code=trust_remote_code
    )
    if student_tokenizer.vocab_size != teacher_tokenizer.vocab_size:
        raise ValueError(
            "Student and teacher tokenizers have different vocab sizes; Support will be added in the future"
        )
    if student_tokenizer.pad_token != teacher_tokenizer.pad_token:
        raise ValueError("Student and teacher tokenizers have different pad tokens")
    del student_tokenizer, teacher_tokenizer


class KnowledgeDistillationRecipeForNextTokenPrediction(TrainFinetuneRecipeForNextTokenPrediction):
    """Fine-tune a student model via knowledge distillation."""

    def setup(self):  # noqa: C901 – same complexity as parent
        """Build student & teacher, dataloaders, optimizers, etc."""
        # Right now, we only support tokenizer compatibility for the same tokenizer.
        # We will add support for different tokenizers in the future.
        _verify_tokenizer_compatibility(self.cfg.get("model", None), self.cfg.get("teacher_model", None))

        # Let the parent class build *everything* for the student first
        super().setup()
        if self.pp_enabled:
            raise ValueError("Pipeline parallelism support will be added in the future for knowledge distillation")

        self._offload_teacher_model = self.cfg.get("offload_teacher_model", False)
        teacher_device = self.dist_env.device if not self._offload_teacher_model else "cpu"

        self.teacher_model = _build_teacher_model(
            cfg_teacher=self.cfg.get("teacher_model", None),
            seed=self.cfg.get("seed", 42),
            has_packed_sequence=self.cfg.get("packed_sequence.packed_sequence_size", 0) > 0,
            device_mesh=self.device_mesh,
            moe_mesh=self.moe_mesh,
            distributed_config=self.distributed_config,
            device=teacher_device,
        )
        logger.info("Teacher Model: " + str(self.teacher_model))
        # KD
        self.kd_loss_fn = _build_kd_loss_fn(self.cfg.get("kd_loss_fn", None))
        self.kd_ratio: float = float(self.cfg.get("kd_ratio", 0.5))
        logger.info("KD Loss config: " + str(self.cfg.get("kd_loss_fn", None)))
        temperature = getattr(self.kd_loss_fn, "temperature", "N/A")
        logger.info(f"Knowledge-distillation enabled: ratio={self.kd_ratio}, T={temperature}")

        # Buffers for logging
        self._kd_loss_buffer = []
        self._ce_loss_buffer = []

    #  Override the forward backward step to inject KD loss
    def _forward_backward_step(
        self,
        idx,
        batch,
        *,
        num_label_tokens,
        num_batches,
        is_train: bool = True,
    ):
        """Override the forward backward step to include knowledge distillation loss."""
        batch = {k: v.to(self.dist_env.device, non_blocking=True) for k, v in batch.items()}
        labels = batch.pop("labels")
        train_ctx, batch = make_cp_batch_and_ctx(self.device_mesh, batch, labels)

        model = self.model_parts[0]
        sync_ctx = (
            get_sync_ctx(
                model,
                idx == num_batches - 1,
                defer_fsdp_grad_sync=getattr(self.distributed_config, "defer_fsdp_grad_sync", True),
            )
            if is_train
            else nullcontext()
        )
        with train_ctx(), sync_ctx:
            # No grad for teacher forward
            with (
                ScopedModuleOffloading(self.teacher_model, enabled=self._offload_teacher_model),
                torch.inference_mode(),
            ):
                teacher_logits = self.teacher_model(**batch)
                teacher_logits = getattr(teacher_logits, "logits", teacher_logits).detach().clone()

            # Student forward
            student_keep_last = isinstance(self.loss_fn, FusedLinearCrossEntropy)
            if student_keep_last:
                # Student forward keeping only last token logits to match loss_fn
                student_out = model(logits_to_keep=1, **batch)
            else:
                student_out = model(**batch)

            student_logits = getattr(student_out, "logits", student_out)  # shape (B, S, V)
            # Cross-entropy loss against true labels (same as parent)
            ce_loss = calculate_loss(
                self.loss_fn,
                logits=student_logits,
                labels=labels,
                model=model,
                hidden_states=student_out.hidden_states[-1] if "hidden_states" in student_out else None,
                num_label_tokens=num_label_tokens,
            )
            # Reminder: kd_loss is normalized by num_label_tokens,
            # which typically is larger than the number of labels in this batch,
            # because it contains the total number of labels for all batches contained
            # in one optimization step (grad_acc_steps = gbs / mbs).
            kd_loss = self.kd_loss_fn(
                student_logits,
                teacher_logits,
                labels,
                num_batch_labels=num_label_tokens,
            )
            local_loss = (1.0 - self.kd_ratio) * ce_loss + self.kd_ratio * kd_loss
            if is_train:
                (local_loss * self._get_dp_group_size(include_cp=True)).backward()
            # return the losses for logging
            detached_local = local_loss.detach().clone()
            return detached_local, kd_loss.detach().clone(), ce_loss.detach().clone()

    def _run_train_optim_step(self, batches, max_grad_norm: Optional[float] = None):
        """Execute a single training step.

        Args:
            batches: List of batches of training data.
            max_grad_norm: Gradient clipping norm. Optional, if None will not clip gradients.
        """

        num_label_tokens = torch.tensor(
            sum((batch["labels"] != -100).sum().item() for batch in batches), dtype=torch.long
        )
        num_label_tokens = self._dp_allreduce(num_label_tokens).item()
        loss_buffer = []

        # number of tokens in the batch, excluding any tail padding.
        num_tokens_in_batch = torch.tensor(
            sum(batch["labels"].numel() - count_tail_padding(batch["labels"]) for batch in batches),
            dtype=torch.long,
        )
        num_tokens_in_batch = self._dp_allreduce(num_tokens_in_batch).item()
        num_batches = len(batches)
        for i, batch in enumerate(batches):
            local_loss, kd_loss, ce_loss = self._forward_backward_step(
                i, batch, num_label_tokens=num_label_tokens, num_batches=num_batches
            )
            loss_buffer.append(local_loss)
            self._ce_loss_buffer.append(ce_loss)
            self._kd_loss_buffer.append(kd_loss)

        grad_norm = 0
        # Clip gradients **after** any rescaling.
        # TODO(@boxiangw): Fix TP gradient clipping
        if max_grad_norm is not None:
            if not self.device_mesh or self.device_mesh["tp"].size() == 1:
                grad_norm = torch.nn.utils.clip_grad_norm_(
                    [p for p in self.model_parts[0].parameters() if p.requires_grad], max_grad_norm
                )
                if hasattr(grad_norm, "full_tensor"):
                    grad_norm = grad_norm.full_tensor()  # collect the summed grad norm across ranks

            if isinstance(grad_norm, torch.Tensor):
                grad_norm = grad_norm.item()

        self.checkpointer.maybe_wait_for_staging()
        for opt in self.optimizer:
            opt.step()
            opt.zero_grad()

        if self.lr_scheduler is not None:
            for scheduler in self.lr_scheduler:
                scheduler.step(1)

        # Precompute FP8 scales
        fp8_config = self.cfg.get("fp8", None)
        if (
            fp8_config is not None
            and fp8_config.get("enabled", False)
            and fp8_config.get("precompute_float8_dynamic_scale_for_fsdp", False)
            and not self.pp_enabled
            and self.device_mesh is not None
            and self.device_mesh["dp_shard"].size() > 1
        ):
            precompute_float8_dynamic_scale_for_fsdp(self.model_parts[0])

        # Note(MegatronFSDP): Need to call these functions for MegatronFSDP if not using latest api
        # self.model_parts[0].install_optimized_model_weights()
        # self.model_parts[0].zero_grad_buffer()

        t = time.perf_counter()
        time_delta = t - self.timestamp
        self.timestamp = t
        tps = num_tokens_in_batch / time_delta
        reporting_loss = torch.sum(torch.stack(loss_buffer))
        reporting_loss = self._dp_allreduce(reporting_loss, include_cp=True)
        reporting_loss = reporting_loss.cpu().item()
        # fix reporting_loss, tps across ranks

        ce_loss = self._dp_allreduce(torch.stack(self._ce_loss_buffer).sum(), include_cp=True).item()
        kd_loss = self._dp_allreduce(torch.stack(self._kd_loss_buffer).sum(), include_cp=True).item()
        # Clear buffers for next step
        self._ce_loss_buffer.clear()
        self._kd_loss_buffer.clear()

        # return reporting_loss, grad_norm, tps, num_tokens_in_batch, num_label_tokens
        return MetricsSample(
            step=self.step_scheduler.step,
            epoch=self.step_scheduler.epoch,
            metrics={
                "loss": reporting_loss,
                "ce_loss": ce_loss,
                "kd_loss": kd_loss,
                "grad_norm": grad_norm,
                "lr": self.optimizer[0].param_groups[0]["lr"],
                "mem": torch.cuda.max_memory_allocated() / 1024**3,
                "tps": tps,
                "tps_per_gpu": tps / max(self._get_dp_group_size(), 1),
                "num_tokens_per_step": num_tokens_in_batch,
                "num_label_tokens": num_label_tokens,
                "kd_ratio": self.kd_ratio,
                "temperature": getattr(self.kd_loss_fn, "temperature", float("nan")),
            },
        )

    @torch.no_grad()
    def _run_validation_epoch(self, val_dataloader):
        """Run one pass over `self.val_dataloader`."""
        if self.pp_enabled:
            logger.warning("Validation is not supported for pipeline parallelism")
            return

        with ScopedRNG(seed=1, ranked=True):
            for mp in self.model_parts:
                mp.eval()

            total_loss = torch.tensor(0.0, dtype=torch.float32, device=self.dist_env.device)
            ce_loss = torch.tensor(0.0, dtype=torch.float32, device=self.dist_env.device)
            kd_loss = torch.tensor(0.0, dtype=torch.float32, device=self.dist_env.device)
            total_num_label_tokens = 0

            for batch in val_dataloader:
                num_label_tokens = (batch["labels"] != -100).sum().item()
                local_loss, _kd_loss, _ce_loss = self._forward_backward_step(
                    0,
                    batch,
                    num_label_tokens=num_label_tokens,
                    num_batches=1,
                    is_train=False,
                )
                total_num_label_tokens += num_label_tokens
                ce_loss += _ce_loss
                kd_loss += _kd_loss
                total_loss += local_loss

        total_loss = self._dp_allreduce(total_loss, include_cp=True).item()
        ce_loss = self._dp_allreduce(ce_loss, include_cp=True).item()
        kd_loss = self._dp_allreduce(kd_loss, include_cp=True).item()
        total_num_label_tokens = self._dp_allreduce(torch.tensor(total_num_label_tokens, dtype=torch.long)).item()

        val_loss = total_loss / max(total_num_label_tokens, 1e-8)
        return MetricsSample(
            step=self.step_scheduler.step,
            epoch=self.step_scheduler.epoch,
            metrics={
                "val_loss": val_loss,
                "ce_loss": ce_loss,
                "kd_loss": kd_loss,
                "lr": self.optimizer[0].param_groups[0]["lr"],
                "num_label_tokens": total_num_label_tokens,
                "mem": torch.cuda.max_memory_allocated() / 1024**3,
            },
        )

    def log_val_metrics(self, val_name, log_data, metric_logger=None):
        if not self.dist_env.is_main or log_data is None:
            return

        if wandb.run is not None:
            wandb.log(log_data.to_dict() | {"val_name": val_name}, step=log_data.step)

        # JSONL validation log
        if not metric_logger is None:
            metric_logger.log(log_data)

        # assumes all model parts' optimizers have the same learning rate
        logging.info(
            "[val] {} | step {} | epoch {} | loss {:.4f} | ce_loss {:.4f} | kd_loss {:.4f} | lr {:.2e} | num_label_tokens {}".format(
                val_name,
                log_data.step,
                log_data.epoch,
                log_data.metrics["val_loss"],
                log_data.metrics["ce_loss"],
                log_data.metrics["kd_loss"],
                log_data.metrics["lr"],
                log_data.metrics["num_label_tokens"],
            )
        )

    def log_train_metrics(self, log_data) -> float:
        """Log metrics to wandb and other loggers.

        Args:
            log_data: MetricsSample object, containing:
                step: int, the current step.
                epoch: int, the current epoch.
                metrics: Dict[str, float], containing:
                    "loss": Training loss.
                    "grad_norm": Grad norm from the training step.
                    "lr": Learning rate.
                    "mem": Memory allocated.
                    "tps": Tokens per second.
                    "tps_per_gpu": Tokens per second per GPU.
                    "num_label_tokens": Number of label tokens.
        """
        if not self.dist_env.is_main:
            return

        # Log to remote services (WandB) according to step_scheduler frequency
        if self.step_scheduler.is_remote_logging_step:
            if wandb.run is not None:
                wandb.log(log_data.to_dict(), step=log_data.step)

        logging.info(
            "step {} | epoch {} | "
            "loss {:.4f} | ce_loss {:.4f} | kd_loss {:.4f} | "
            "lr {:.2e} | tps {:.2f} | kd_ratio {:.2f} | temperature {:.2f}".format(
                log_data.step,
                log_data.epoch,
                log_data.metrics["loss"],
                log_data.metrics["ce_loss"],
                log_data.metrics["kd_loss"],
                log_data.metrics["lr"],
                log_data.metrics["tps"],
                log_data.metrics["kd_ratio"],
                log_data.metrics["temperature"],
            )
        )
        torch.cuda.reset_peak_memory_stats()


# Entry point
def main(config_path="examples/llm_kd/llama3_2/llama3_2_1b_kd.yaml"):
    """Run the KD recipe from CLI or directly."""
    cfg = parse_args_and_load_config(config_path)
    trainer = KnowledgeDistillationRecipeForNextTokenPrediction(cfg)
    trainer.setup()
    trainer.run_train_validation_loop()


if __name__ == "__main__":  # pragma: no cover
    main()
